import json
import pathlib

import jax
import jax.numpy as jnp
import equinox as eqx

import exciting_environments as excenvs
from dmpe.models.models import NeuralEulerODE


@eqx.filter_jit
def simulate_ahead(
    model: NeuralEulerODE,
    init_obs: jax.Array,
    actions: jax.Array,
    tau: float,
) -> jax.Array:
    """Uses the given model to look ahead and simulate future observations.

    Args:
        model (eqx.Module): The model to use in the simulation
        init_obs (jax.Array): The initial observation from which to start the simulation
        actions (jax.Array): The actions to apply in each step of the simulation, the length
            of the first dimension of this array determines the length of the output.
        tau (float): The sampling time for the model

    Returns:
        observations (jax.Array): The predicted observations. The shape of this is given as
            (n_actions + 1, obs_dim). That is because the first observation is
            already given through the initial observation
    """
    return model(init_obs, actions, tau)


@eqx.filter_jit
def simulate_ahead_with_env(
    env: excenvs.CoreEnvironment,
    init_obs: jax.Array,
    init_state: excenvs.CoreEnvironment.State,
    actions: jax.Array,
) -> jax.Array:
    """Uses the given environment to look ahead and simulate future observations.
    This is used to have perfect predictions.

    Args:
        env (excenvs.CoreEnvironment): The env used for the simulation
        init_obs (jax.Array): The initial observation from which to start the simulation
        init_state (excenvs.CoreEnvironment.State): The initial state from which to start
            the simulation
        actions (jax.Array): The actions to apply in each step of the simulation, the length
            of the first dimension of this array determine the length of the output.

    Returns:
        observations (jax.Array): The simulated observations. The shape of this is given as
            (n_actions + 1, obs_dim). That is because the first observation is already given
            through the initial observation
    """

    def body_fun(carry, action):
        obs, state = carry

        obs, state = env.step(state, action, env.env_properties)
        return (obs, state), obs

    (_, last_state), observations = jax.lax.scan(body_fun, (init_obs, init_state), actions)
    observations = jnp.concatenate([init_obs[None, :], observations], axis=0)

    return observations, last_state


def save_model(filename: str | pathlib.Path, hyperparams: dict, model: eqx.Module):
    """Store the given model + hyperparameters at the specified file path."""
    with open(filename, "wb") as f:
        hyperparam_str = json.dumps(hyperparams)
        f.write((hyperparam_str + "\n").encode())
        eqx.tree_serialise_leaves(f, model)


def load_model(filename: str, model_class: eqx.Module):
    """Load the model and hyperparameters from the specified file path.

    Note that the hyperparameters must be sufficient to initialized the given
    model class.
    """
    with open(filename, "rb") as f:
        hyperparams = json.loads(f.readline().decode())
        hyperparams["key"] = jnp.array(hyperparams["key"], dtype=jnp.uint32)
        model = model_class(**hyperparams)
        return eqx.tree_deserialise_leaves(f, model)



"""System excitation algorithms from related work."""

from typing import Callable

from tqdm import tqdm
import matplotlib.pyplot as plt
import numpy as np
from pymoo.core.mixed import MixedVariableGA, MixedVariableDuplicateElimination

from dmpe.related_work.excitation_utils import (
    optimize_continuous_aprbs,
    optimize_permutation_aprbs,
    latin_hypercube_sampling,
)

import exciting_environments as exenvs
from dmpe.evaluation.plotting_utils import plot_sequence
from dmpe.related_work.mixed_GA import MixedVariableMating, MixedVariableSampling


def excite_with_GOATS(
    n_amplitudes: np.ndarray,
    env,
    bounds_duration: tuple,
    population_size: int,
    n_generations: int,
    featurize: Callable,
    rng: np.random.Generator,
    compress_data: bool,
    compression_target_N: int,
    rho_act: float,
    rho_obs: float,
    penalty_order: int,
    compression_dist_th: float,
    compression_feat_dim: int,
    verbose: bool = True,
):
    """System excitation using the GOATs algorithm from [Smits2024].

    The optimization metric used here is audze eglais as it is described in [Smits2024]
    and it is optimized with a genetic algorithm.

    Args:
        n_amplitudes: The number of amplitudes to sample with Latin Hypercube Sampling
            and optimize w.r.t. the system excitation task
        env: The environment/system or model to excite
        bounds_duration: The upper and lower bound on the duration of a single amplitude
            as a tuple
        population_size: The number of individuals in the population of the GA
        n_generations: The number of generations of the GA
        featurize: Featurization of the observations before computation of the metric. If
            this is not necessary for the environment/system, pass the identity function
        rng: The random number generator for the genetic algorithm and LHS
        verbose: Whether the genetic algorithm print out its optimization progress

    Returns:
        observations: The observations gathered from the system
        actions: The actions applied to the system
    """

    obs, env_state = env.reset(env.env_properties)

    opt_algorithm = MixedVariableGA(
        pop_size=population_size,
        sampling=MixedVariableSampling(),
        mating=MixedVariableMating(eliminate_duplicates=MixedVariableDuplicateElimination()),
    )

    observations, actions, _ = optimize_permutation_aprbs(
        opt_algorithm,
        amplitudes=latin_hypercube_sampling(d=env.action_dim, n=n_amplitudes, rng=rng),
        env=env,
        obs=obs,
        env_state=env_state,
        bounds_duration=bounds_duration,
        n_generations=n_generations,
        featurize=featurize,
        rng=rng,
        verbose=verbose,
        starting_observations=None,
        starting_actions=None,
        compress_data=compress_data,
        compression_target_N=compression_target_N,
        rho_act=rho_act,
        rho_obs=rho_obs,
        penalty_order=penalty_order,
        compression_dist_th=compression_dist_th,
        compression_feat_dim=compression_feat_dim,
        share_of_current_sequence=1,
    )

    return observations, actions


def generate_amplitude_groups(n_amplitudes: int, n_amplitude_groups: int, rng: np.random.Generator) -> np.ndarray:
    """
    Generate amplitude groups from a given number of amplitudes where each group in itself is
    "space-filling" in the sense that it is uniform on the interval [-1, 1].

    Args:
        n_amplitudes (int): The total number of amplitudes.
        n_amplitude_groups (int): The number of amplitude groups to generate.
        rng (np.random.Generator): The random number generator.

    Returns:
        amplitude_groups (np.ndarray): An array of amplitude groups, where each group contains a subset of amplitudes.

    Raises:
        AssertionError: If n_amplitudes is not divisible by n_amplitude_groups.

    """
    assert n_amplitudes % n_amplitude_groups == 0
    all_amplitudes = np.linspace(-1, 1, n_amplitudes)

    amplitude_groups = [[] for _ in range(n_amplitude_groups)]

    for idx, amplitude in enumerate(all_amplitudes):
        i = idx % n_amplitude_groups
        amplitude_groups[i].append(amplitude)

    rng.shuffle(amplitude_groups)
    amplitude_groups = np.array(amplitude_groups)

    for amplitude_group in amplitude_groups:
        rng.shuffle(amplitude_group)

    return amplitude_groups


def excite_with_sGOATS(
    n_amplitudes: np.ndarray,
    n_amplitude_groups: int,
    reuse_observations: bool,
    env: exenvs.CoreEnvironment,
    bounds_duration: tuple,
    population_size: int,
    n_generations: int,
    featurize: Callable,
    rng: np.random.Generator,
    compress_data: bool,
    compression_target_N: int,
    rho_act: float,
    rho_obs: float,
    penalty_order: int,
    compression_dist_th: float,
    compression_feat_dim: int,
    verbose: bool = True,
    plot_every_subsequence: bool = True,
):
    """System excitation using the sGOATs algorithm from [Smits2024].

    The optimization metric used here is MCUDSA as it is described in [Smits2024]
    and it is optimized with a genetic algorithm. It is similar to the GOATs algorithm
    with the major difference that the precomputed amplitude levels are not all used
    at once but only a subset of the amplitude levels is optimized at a time.

    Args:
        n_amplitudes: The number of amplitudes to sample with Latin Hypercube Sampling
            and optimize w.r.t. the system excitation task
        n_amplitude_groups: Decides in how many groups the amplitude levels are split
            for optimization
        reuse_observations: Whether to reuse the observations from previous amplitude
            groups in the metric computations of the following groups
        all_observations: A list to which the gathered observations are appended. When
            you want to use the algorithm as it is intended, this should be an empty
            list.
        all_actions: A list to which the applied actions are appended. When you want
            to use the algorithm as it is intended, this should be an empty list.
        env: The environment/system or model to excite
        bounds_duration: The upper and lower bound on the duration of a single amplitude
            as a tuple
        population_size: The number of individuals in the population of the GA
        n_generations: The number of generations of the GA
        featurize: Featurization of the observations before computation of the metric. If
            this is not necessary for the environment/system, pass the identity function
        rng: The rng for the genetic algorithm.
        verbose: Whether the genetic algorithm print out its optimization progress

    Returns:
        all_observations: The finished observations list
        all_actions: The finished actions list
    """

    all_observations = []
    all_actions = []

    opt_algorithm = MixedVariableGA(
        pop_size=population_size,
        sampling=MixedVariableSampling(),
        mating=MixedVariableMating(eliminate_duplicates=MixedVariableDuplicateElimination()),
    )

    obs, env_state = env.reset(env.env_properties)

    amplitude_groups = generate_amplitude_groups(
        n_amplitudes=n_amplitudes, n_amplitude_groups=n_amplitude_groups, rng=rng
    )

    for idx, amplitudes in enumerate(tqdm(amplitude_groups)):

        if len(all_observations) > 0 and reuse_observations:
            starting_observations = np.concatenate(all_observations)
        else:
            starting_observations = None

        if len(all_actions) > 0 and reuse_observations:
            starting_actions = np.concatenate(all_actions)
        else:
            starting_actions = None

        observations, actions, last_env_state = optimize_permutation_aprbs(
            opt_algorithm=opt_algorithm,
            amplitudes=amplitudes,
            env=env,
            obs=obs,
            env_state=env_state,
            bounds_duration=bounds_duration,
            n_generations=n_generations,
            featurize=featurize,
            rng=rng,
            verbose=verbose,
            starting_observations=starting_observations,
            starting_actions=starting_actions,
            compress_data=compress_data,
            compression_target_N=compression_target_N,
            rho_act=rho_act,
            rho_obs=rho_obs,
            penalty_order=penalty_order,
            compression_dist_th=compression_dist_th,
            compression_feat_dim=compression_feat_dim,
            share_of_current_sequence=1 / (idx + 1),
        )

        # update obs and env_state as the starting point for the next amplitude group
        obs = observations[-1, :]
        env_state = last_env_state

        # save optimized actions and resulting observations
        all_observations.append(observations[:-1, :])
        all_actions.append(actions)

        if plot_every_subsequence:
            fig, axs = plot_sequence(
                observations=np.concatenate(all_observations),
                actions=np.concatenate(all_actions)[:-1, ...],
                tau=env.tau,
                obs_labels=env.obs_description,
                action_labels=env.action_description,
            )
            plt.show()

    return all_observations, all_actions


def excite_with_iGOATS(
    n_time_steps,
    env,
    prediction_horizon,
    application_horizon,
    bounds_amplitude,
    bounds_duration,
    population_size,
    n_generations,
    featurize,
    rng,
    compress_data,
    compression_target_N,
    compression_feat_dim,
    compression_dist_th,
    penalty_function,
    plot_subsequences=False,
):
    """System excitation using the iGOATs algorithm from [Smits2024]."""

    assert application_horizon <= prediction_horizon
    obs, env_state = env.reset(env.env_properties)

    all_actions = []
    all_observations = []

    opt_algorithm = MixedVariableGA(
        pop_size=population_size,
        sampling=MixedVariableSampling(),
        mating=MixedVariableMating(eliminate_duplicates=MixedVariableDuplicateElimination()),
    )

    pbar = tqdm(total=n_time_steps)
    while True:

        if len(all_observations) > 0 and len(all_actions) > 0:
            starting_observations = np.concatenate(all_observations)
            starting_actions = np.concatenate(all_actions)

        else:
            starting_observations = None
            starting_actions = None

        new_observations, new_actions, env_state = optimize_continuous_aprbs(
            opt_algorithm,
            prediction_horizon=prediction_horizon,
            application_horizon=application_horizon,
            env=env,
            obs=obs,
            env_state=env_state,
            bounds_amplitude=bounds_amplitude,
            bounds_duration=bounds_duration,
            n_generations=n_generations,
            featurize=featurize,
            rng=rng,
            starting_observations=starting_observations,
            starting_actions=starting_actions,
            compress_data=compress_data,
            compression_target_N=compression_target_N,
            compression_dist_th=compression_dist_th,
            compression_feat_dim=compression_feat_dim,
            penalty_function=penalty_function,
        )

        obs = new_observations[-1]

        all_observations.append(new_observations[:-1, :])
        all_actions.append(new_actions)

        if plot_subsequences:
            fig, axs = plot_sequence(
                observations=np.concatenate(all_observations),
                actions=np.concatenate(all_actions),
                tau=env.tau,
                obs_labels=env.obs_description,
                action_labels=env.action_description,
            )
            plt.show()

        pbar.update(new_actions.shape[0])

        if np.concatenate(all_observations).shape[0] >= n_time_steps:
            break

    pbar.close()

    return all_observations, all_actions



import jax
import jax.numpy as jnp


@jax.jit
def KLDLoss(p: jax.Array, q: jax.Array) -> jax.Array:
    """Computes the sample Kullback-Leibler divergence (KLD) between two inputs.

    The last dim of the input needs to be of length 1. The summation occurs along the second to
    last dimension. All dimensions before that are kept as they are. Overall the shape of the
    two inputs must be identical.
    A small constant is added to the inputs to ensure numerical computability for inputs with
    values approaching zero.
    """
    assert p.shape == q.shape, "The two inputs need to be of the same shape."
    assert p.shape[-1] == q.shape[-1] == 1, "Last dim needs to be of length 1 for PDFs"

    eps = 1e-12

    kld = (p + eps) * jnp.log((p + eps) / (q + eps))
    return jnp.squeeze(jnp.sum(kld, axis=-2))


@jax.jit
def JSDLoss(p: jax.Array, q: jax.Array) -> jax.Array:
    """Computes the sample Jensen-Shannon divergence (JSD) between two inputs.

    The last dim of the input needs to be of length 1. The summation occurs along the second to
    last dimension. All dimensions before that are kept as they are. Overall the shape of the
    two inputs must be identical.
    """
    assert p.shape == q.shape, "The two inputs need to be of the same shape."
    assert p.shape[-1] == q.shape[-1] == 1, "Last dim needs to be of length 1 for PDFs"

    m = (p + q) / 2
    return jnp.squeeze((KLDLoss(p, m) + KLDLoss(q, m)) / 2)


def MNNS_without_penalty(data_points: jax.Array, new_data_points: jax.Array) -> jax.Array:
    """From [Smits2024: "Space-filling optimized excitation signals for nonlinear system
    identification of dynamic processes of a diesel engine", V. Smits et al., 2024].

    Implementation inspired by https://github.com/google/jax/discussions/9813

    TODO: Not sure about this penalty. Seems difficult to use for continuous action-spaces?
    They used quantized amplitude levels in their implementation.
    """
    L = new_data_points.shape[0]
    distance_matrix = jnp.linalg.norm(data_points[:, None, :] - new_data_points[None, ...], axis=-1)
    minimal_distances = jnp.min(distance_matrix, axis=0)
    return -jnp.sum(minimal_distances) / L


def audze_eglais(data_points: jax.Array, eps: float = 0.001) -> jax.Array:
    """From [Smits2024]. The maximin-design penalizes points that are too close to each other in
    the data point distribution.
    """
    N = data_points.shape[0]
    distance_matrix = jnp.linalg.norm(data_points[:, None, :] - data_points[None, ...], axis=-1)
    distances = distance_matrix[jax.numpy.triu_indices(N, k=1)]

    return 2 / (N * (N - 1)) * jnp.sum(1 / (distances**2 + eps))


@jax.jit
def MC_uniform_sampling_distribution_approximation(data_points: jax.Array, support_points: jax.Array) -> jax.Array:
    """From [Smits2024]. The minimax-design aims to minimize the distances of the data points to
    the support points.

    The loss is computed by considering the distance to the closest datapoint for each support
    point.
    """
    M = support_points.shape[0]
    distance_matrix = jnp.linalg.norm(data_points[:, None, :] - support_points[None, ...], axis=-1)
    minimal_distances = jnp.min(distance_matrix, axis=0)

    return jnp.sum(minimal_distances) / M


def blockwise_mcudsa(
    data_points: jax.Array,
    support_points: jax.Array,
    block_size: int = 1000,
) -> jax.Array:
    """Block-wise implementation of MCUDSA.

    Splits the support points into blocks with 'block_size' elements to reduce the amount of memory
    necessary for the computation.
    """

    M = support_points.shape[0]
    value = jnp.zeros(1)

    for m in range(0, M, block_size):
        end = min(m + block_size, M)  # next block or until the end
        value = value + (
            MC_uniform_sampling_distribution_approximation(
                data_points=data_points,
                support_points=support_points[m:end],
            )
            * (end - m)  # denormalizing mean inside loss computation
            / M
        )

    return value


@jax.jit
def kiss_space_filling_cost(
    data_points: jax.Array,
    support_points: jax.Array,
    variances: jax.Array,
    eps: float = 1e-16,
) -> jax.Array:
    """From [Kiss2024: "Space-Filling Input Design for Nonlinear State-Space Identification",
    M. Kiss et al., 2024].

    Slightly modified to use the mean instead of the sum in the denominator.
    The goal is to have the same metric value for identical data distributions with different number
    of data points.
    """
    difference = data_points[None, ...] - support_points[:, None, :]
    exponent = -0.5 * jnp.sum(difference**2 * 1 / variances, axis=-1)

    denominator = eps + jnp.mean(jnp.exp(exponent), axis=-1)

    return jnp.mean(1 / denominator, axis=0)


def blockwise_ksfc(
    data_points: jax.Array,
    support_points: jax.Array,
    variances: jax.Array,
    eps: float = 1e-16,
    block_size: int = 1000,
) -> jax.Array:
    """Block-wise implementation of KSFC [Kiss2024].

    Splits the support points into blocks with 'block_size' elements to reduce the amount of memory
    necessary for the computation.
    """

    M = support_points.shape[0]
    value = jnp.zeros(1)

    for m in range(0, M, block_size):
        end = min(m + block_size, M)  # next block or until the end
        value = value + (
            kiss_space_filling_cost(
                data_points=data_points,
                support_points=support_points[m:end],
                variances=variances,
                eps=eps,
            )
            * (end - m)  # denormalizing mean inside loss computation
            / M
        )

    return value



from functools import partial
import json
import time
import argparse
import datetime
import pathlib

import numpy as np
import jax
import jax.numpy as jnp

from dmpe.utils.density_estimation import build_grid
from dmpe.utils.env_utils.foc_pi import ClassicController
from dmpe.evaluation.scripts.pmsm.eval_dmpe import setup_env, TARGETED_DATA_PATH


@partial(jax.jit, static_argnums=(0, 1))
def run_pi_experiment(env, pi, references_norm, init_obs, init_state, init_pi_state):
    """Rollout the experiment by following the given references with the PI controller."""

    def body_fun(carry, reference_norm):
        obs, state, pi_state = carry

        currents_norm = obs[None]
        eps = state.physical_state.epsilon * jnp.ones((1, 1))
        reference_norm = reference_norm[None, ...]
        pi_obs = jnp.concatenate([currents_norm, eps, reference_norm], axis=-1)

        action, next_pi_state = pi(pi_obs, pi_state)

        next_obs, next_state = env.step(state, jnp.squeeze(action), env.env_properties)
        return (next_obs, next_state, next_pi_state), jnp.array([jnp.squeeze(obs), jnp.squeeze(action)])

    (_, _, _), data = jax.lax.scan(body_fun, (init_obs, init_state, init_pi_state), references_norm)
    observations = data[:, 0, :]
    actions = data[:, 1, :]

    return observations, actions


def subsample_references(references, target_N):
    M = references.shape[0]

    x_old = jnp.linspace(0, M - 1, M)
    x_new = jnp.linspace(0, M - 1, target_N)

    def linear_interp(x, xp, fp):
        idx = jnp.searchsorted(xp, x) - 1
        idx = jnp.clip(idx, 0, len(fp) - 2)
        x0, x1 = xp[idx], xp[idx + 1]
        y0, y1 = fp[idx], fp[idx + 1]
        return y0 + (y1 - y0) * (x - x0)[:, None] / (x1 - x0)[:, None]

    return linear_interp(x_new, x_old, references)


def induced_voltage_constr(z_g, env, w):
    """Compute voltage constraint violations."""
    r_s = env.env_properties.static_params.r_s

    i_d_normalizer = env.env_properties.physical_normalizations.i_d
    i_q_normalizer = env.env_properties.physical_normalizations.i_q
    physical_i_d = i_d_normalizer.denormalize(z_g[0])
    physical_i_q = i_q_normalizer.denormalize(z_g[1])

    psid = env.LUT_interpolators["Psi_d"](jnp.array([physical_i_d, physical_i_q]))
    psiq = env.LUT_interpolators["Psi_q"](jnp.array([physical_i_d, physical_i_q]))

    ud = r_s * physical_i_d - w * psiq
    uq = r_s * physical_i_q + w * psid

    u = jnp.sqrt(ud**2 + uq**2)

    return jnp.squeeze(jax.nn.relu(u - 400 / jnp.sqrt(3)))


def filter_voltage_constraints(env, rpm, references):
    """Filter out references that violate the voltage constraints."""
    penalty_values = jax.vmap(induced_voltage_constr, in_axes=(0, None, None))(
        references, env, rpm * env.env_properties.static_params.p * 2 * jnp.pi / 60
    )

    valid_points = penalty_values == 0
    filtered_references = references[jnp.where(valid_points == True)]

    return filtered_references


def setup_references(env, rpm, points_per_dim, penalty_function, n_timesteps=15_000):
    """Setup a reference trajectory to follow with the PI controller."""
    references = build_grid(2, low=-0.95, high=0.95, points_per_dim=points_per_dim)
    references = jnp.flip(references, axis=1)
    references = jnp.flip(references, axis=0)

    references = references.reshape(points_per_dim, points_per_dim, 2)
    references = np.array(references)
    for k in range(points_per_dim):
        if k % 2 == 0:
            references[k] = np.flip(references[k], axis=0)
    references = references.reshape(points_per_dim**2, 2)
    references = references[int(points_per_dim / 2) :]

    references = jnp.array(references)

    references_norm = references[:, None, :].repeat(5, axis=1).reshape(-1, 2)
    # references_norm = references[:, None, :].repeat(500, axis=1).reshape(-1, 2)

    def _filter_valid_points(data_points, penalty_function):

        valid_points_bool = jax.vmap(penalty_function, in_axes=0)(data_points) == 0
        return data_points[jnp.where(valid_points_bool == True)]

    filter_valid_points = lambda observations: _filter_valid_points(
        observations, penalty_function=lambda x: penalty_function(x, None)
    )

    references_norm = filter_valid_points(references_norm)
    references_norm = filter_voltage_constraints(env, rpm, references_norm)
    references_norm = subsample_references(references_norm, n_timesteps)
    return references_norm


def run_experiment(rpm):
    print(
        "Running experiment with PI controller",
        f"on the PMSM with {rpm} rpm.",
    )

    # Check that the targeted data folder actually exist:
    results_path = TARGETED_DATA_PATH / pathlib.Path("heuristics") / pathlib.Path("current_plane_sweep")
    print(f"Results will be written to: '{results_path}'.")
    assert results_path.exists(), (
        f"The expected results path '{results_path}' does not seem to exist. Please create the necessary file structure "
        + "or adapt the path."
    )

    points_per_dim = 100

    env, penalty_function = setup_env(rpm)

    pi = ClassicController(
        motor=env, rpm=rpm, saturated=env.env_properties.saturated, a=4, decoupling=True, tau=env.tau
    )

    references = setup_references(env, rpm, points_per_dim, penalty_function)

    init_obs, init_state = env.reset(env.env_properties)
    init_pi_state = pi.reset(1)

    start = time.time()
    observations, actions = run_pi_experiment(env, pi, references, init_obs, init_state, init_pi_state)
    end = time.time()
    print("computation_time:", round(end - start, 4), "s")

    # experiment finished, save results
    file_name = datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")

    with open(results_path / pathlib.Path(f"data_rpm_{rpm}_{file_name}.json"), "w") as fp:
        json.dump(dict(observations=observations.tolist(), actions=actions.tolist()), fp)

    jax.clear_caches()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Run random walk on the PMSM environment.")
    parser.add_argument("--gpu_id", type=int, default=0, help="GPU id to use.")
    args = parser.parse_args()

    gpus = jax.devices()
    jax.config.update("jax_default_device", gpus[args.gpu_id])

    rpms = [0, 3000, 5000, 7000, 9000]

    for rpm in rpms:
        # no need for a seed since the experiment is deterministic
        run_experiment(rpm)



from typing import Callable
from functools import partial

import jax
import jax.numpy as jnp
import optax
import equinox as eqx

import exciting_environments as excenvs

from dmpe.models.model_utils import simulate_ahead, simulate_ahead_with_env
from dmpe.utils.density_estimation import (
    DensityEstimate,
    update_density_estimate_single_observation,
    update_density_estimate_multiple_observations,
)
from dmpe.utils.metrics import JSDLoss


def soft_penalty(a: jax.Array, a_max: float = 1.0, penalty_order: int = 2):
    """Computes penalty for the given input. Assumes symmetry in all dimensions.

    Args:
        a (jax.Array): The input array
        a_max (float): The maximum absolute value allowed for the input
        penalty_order (int): The order of the penalty term (default is 2)

    Returns:
        penalty (jax.Array): Resulting penalty value
    """
    penalties = jax.nn.relu(jnp.abs(a) - a_max)
    penalties = penalties**penalty_order

    penalty = jnp.sum(penalties, axis=(-2, -1))
    return jnp.squeeze(penalty)


@eqx.filter_jit
def loss_function(
    model: eqx.Module,
    init_obs: jax.Array,
    init_state: excenvs.CoreEnvironment.State,
    actions: jax.Array,
    density_estimate: DensityEstimate,
    tau: float,
    consider_action_distribution: bool,
    target_distribution: jax.Array,
    penalty_function: Callable,
) -> jax.Array:
    """Predicts a trajectory based on the given actions and the model and computes the
    corresponding loss value based on the Jensen-Shannon divergence loss.

    Args:
        model (eqx.Module): The model to use for the prediction
        init_obs (jax.Array): The initial observation from which to start the simulation
        init_state (excenvs.CoreEnvironment.State): The initial state from which to start
            the simulation (only actually used when using the environment for simulation)
        actions (jax.Array): The actions to apply in each step of the simulation, the length
            of the first dimension of this array determine the length of the output.
        density_estimate (DensityEstimate): The current estimate of the data density
        tau (float): The sampling time for the model
        consider_action_distribution (bool): Whether to consider the action as part of the
            feature space
        target_distribution (jax.Array): The goal distribution of the data. The JSD loss is
            computed w.r.t. this distribution
        penalty_function (Callable): A function that computes a penalty term for the observations
            and actions. This is done to penalize overstepping of the physical constraints of the
            system
    """
    if isinstance(model, eqx.Module):
        observations = simulate_ahead(model=model, init_obs=init_obs, actions=actions, tau=tau)
    else:
        observations, _ = simulate_ahead_with_env(env=model, init_obs=init_obs, init_state=init_state, actions=actions)

    if consider_action_distribution:
        predicted_density_estimate = update_density_estimate_multiple_observations(
            density_estimate, jnp.concatenate([observations[0:-1, :], actions], axis=-1)  # observations
        )
    else:
        predicted_density_estimate = update_density_estimate_multiple_observations(density_estimate, observations)

    loss = JSDLoss(
        p=predicted_density_estimate.p / jnp.sum(predicted_density_estimate.p),
        q=target_distribution / jnp.sum(target_distribution),
    )
    penalty_terms = penalty_function(observations, actions)
    return loss + penalty_terms


@eqx.filter_jit
def optimize_actions(
    loss_function: Callable,
    grad_loss_function: Callable,
    proposed_actions: jax.Array,
    model: eqx.Module,
    optimizer: optax.GradientTransformation | optax.GradientTransformationExtraArgs,
    init_obs: jax.Array,
    init_state: excenvs.CoreEnvironment.State,
    density_estimate: DensityEstimate,
    n_opt_steps: int,
    tau: float,
    consider_action_distribution: bool,
    target_distribution: jax.Array,
    penalty_function: Callable,
):
    """Uses the model to compute the effect of actions onto the observation trajectory to
    optimize the actions w.r.t. the given (gradient of the) loss function.

    Args:
        loss_function (Callable): The loss function to optimize
        grad_loss_function (Callable): The gradient of the loss function w.r.t. the actions
        proposed_actions (jax.Array): The initial proposed actions to optimize
        model (eqx.Module): The model to use for the prediction
        optimizer (optax.GradientTransformation): The optimizer to use for the optimization
        init_obs (jax.Array): The initial observation from which to start the simulation
        init_state (excenvs.CoreEnvironment.State): The initial state from which to start
            the simulation
        density_estimate (DensityEstimate): The momentary estimate of the data density
        n_opt_steps (int): Number of SGD steps per iteration
        tau (float): The sampling time for the model and system
        consider_action_distribution (bool): Whether to consider the action as part of the
            feature space
        target_distribution (jax.Array): The goal distribution of the data. The JSD loss is
            computed w.r.t. this distribution
        penalty_function (Callable): A function that computes a penalty term for the observations
            and actions. This is done to penalize overstepping of the physical constraints of the
            system
    Returns:
        proposed_actions (jax.Array): The optimized actions
        loss (jax.Array): The final loss value
    """
    opt_state = optimizer.init(proposed_actions)

    def body_fun(i, carry):
        proposed_actions, opt_state = carry
        value, grad = grad_loss_function(
            model,
            init_obs,
            init_state,
            proposed_actions,
            density_estimate,
            tau,
            consider_action_distribution,
            target_distribution,
            penalty_function,
        )

        if isinstance(optimizer, optax._src.base.GradientTransformationExtraArgs):

            opt_loss_fn = lambda proposed_actions: loss_function(
                model,
                init_obs,
                init_state,
                proposed_actions,
                density_estimate,
                tau,
                consider_action_distribution,
                target_distribution,
                penalty_function,
            )

            updates, opt_state = optimizer.update(
                grad, opt_state, proposed_actions, grad=grad, value=value, value_fn=opt_loss_fn
            )
        else:
            updates, opt_state = optimizer.update(grad, opt_state, proposed_actions)
        proposed_actions = optax.apply_updates(proposed_actions, updates)

        # proposed_actions = proposed_actions - lr * grad

        return (proposed_actions, opt_state)

    proposed_actions, _ = jax.lax.fori_loop(0, n_opt_steps, body_fun, (proposed_actions, opt_state))

    loss = loss_function(
        model,
        init_obs,
        init_state,
        proposed_actions,
        density_estimate,
        tau,
        consider_action_distribution,
        target_distribution,
        penalty_function,
    )

    return proposed_actions, loss


def optimize_actions_multistart(
    loss_function: Callable,
    grad_loss_function: Callable,
    all_proposed_actions: jax.Array,
    model: eqx.Module,
    optimizer: optax.GradientTransformation | optax.GradientTransformationExtraArgs,
    init_obs: jax.Array,
    init_state: excenvs.CoreEnvironment.State,
    density_estimate: DensityEstimate,
    n_opt_steps: int,
    tau: float,
    consider_action_distribution: bool,
    target_distribution: jax.Array,
    penalty_function: Callable,
):
    """Parallelizes the action optimization function w.r.t. the proposed actions. Thereby,
    enables the use of multistart optimization. Only the result with the best final loss
    value is returned from this function.

    This is done since the gradient based optimization methods employed to solve the
    excitation optimization problem can get stuck in local optima. The multistart aims to
    decrease the probability of this occurring.

    Args:
        loss_function (Callable): The loss function to optimize
        grad_loss_function (Callable): The gradient of the loss function w.r.t. the actions
        all_proposed_actions (jax.Array): The initial proposed actions to optimize. Note that
            compared to the 'optimize_actions' function, 'all_proposed_actions' contains an
            extra dimension for the multistart optimization. Shape: (n_starts, n_sim_steps, input_dim).
        model (eqx.Module): The model to use for the prediction
        optimizer (optax.GradientTransformation): The optimizer to use for the optimization
        init_obs (jax.Array): The initial observation from which to start the simulation
        init_state (excenvs.CoreEnvironment.State): The initial state from which to start
            the simulation
        density_estimate (DensityEstimate): The momentary estimate of the data density
        n_opt_steps (int): Number of SGD steps per iteration
        tau (float): The sampling time for the model and system
        consider_action_distribution (bool): Whether to consider the action as part of the
            feature space
        target_distribution (jax.Array): The goal distribution of the data. The JSD loss is
            computed w.r.t. this distribution
        penalty_function (Callable): A function that computes a penalty term for the observations
            and actions. This is done to penalize overstepping of the physical constraints of the
            system
    Returns:
        proposed_actions (jax.Array): The optimized actions
        loss (jax.Array): The final loss value
    """
    assert all_proposed_actions.ndim == 3, "proposed_actions must have shape (n_starts, n_sim_steps, action_dim)"

    all_optimized_actions, all_losses = jax.vmap(
        optimize_actions,
        in_axes=(None, None, 0, None, None, None, None, None, None, None, None, None, None),
    )(
        loss_function,
        grad_loss_function,
        all_proposed_actions,
        model,
        optimizer,
        init_obs,
        init_state,
        density_estimate,
        n_opt_steps,
        tau,
        consider_action_distribution,
        target_distribution,
        penalty_function,
    )

    best_idx = jnp.argmin(all_losses)
    return all_optimized_actions[best_idx], all_losses[best_idx]


class Exciter(eqx.Module):
    """A class that carries the necessary tools for excitation input computations.

    Args:
        start_optimizing (int): When to start optimizing actions
        loss_function (Callable): The loss function to optimize
        grad_loss_function: The gradient of the loss function w.r.t. the actions as
            a callable function
        excitation_optimizer (optax.GradientTransformation): The optimizer for the
            excitation input computation
        tau (float): The time step length of the simulation
        n_opt_steps (int): Number of SGD steps per iteration
        consider_action_distribution (bool): Whether to consider the action as part of the
            feature vector
        target_distribution (jax.Array): The targeted distribution for the data density
        penalty_function (Callable): A function that computes a penalty term for the observations
            and actions. This is done to penalize overstepping of the physical constraints of the
            system
        clip_action (True): Whether to clip the actions to the action space limits
        n_starts (int): Number of parallel starts for the optimization
        reuse_proposed_actions (True): Whether to reuse the proposed actions from the previous
            iteration as the initial guess for the next step (it is only used for one of the
            n_starts)
    """

    start_optimizing: int
    loss_function: Callable
    grad_loss_function: Callable
    excitation_optimizer: optax.GradientTransformation
    tau: float
    n_opt_steps: int
    consider_action_distribution: bool
    target_distribution: jax.Array
    penalty_function: Callable
    clip_action: bool
    n_starts: int
    reuse_proposed_actions: bool

    @eqx.filter_jit
    def choose_action(
        self,
        obs: jax.Array,
        state: excenvs.CoreEnvironment.State,
        model: eqx.Module,
        density_estimate: DensityEstimate,
        proposed_actions: jax.Array,
        expl_key: jax.random.PRNGKey,
    ) -> tuple[jax.Array, jax.Array, DensityEstimate]:
        """Chooses the next action to take, updates the density estimate and
        proposes future actions.

        Args:
            obs (jax.Array): The momentary observations from which to start
            state (excenvs.CoreEnvironment.State): The momentary state from which to start
            model (eqx.Module): The momentary model of the environment used for the prediction
            density_estimate (DensityEstimate): The current estimate of the data density without
                the current step k
            proposed_actions (jax.Array): An initial proposition of actions to take

        Returns:
            action (jax.Array): The chosen action
            next_proposed_actions (jax.Array): An initial proposition for future actions
            density_estimate (DensityEstimate): The updated density estimate now incorporating
                the current step k
        """

        if self.reuse_proposed_actions:
            n_random_starts = self.n_starts - 1

        expl_key, new_proposed_actions_key, expl_action_key = jax.random.split(expl_key, 3)

        if n_random_starts > 0:
            random_proposed_actions = jax.random.uniform(
                key=new_proposed_actions_key, shape=(n_random_starts, *proposed_actions.shape), minval=-1, maxval=1
            )

            if self.reuse_proposed_actions:
                all_proposed_actions = jnp.concatenate([proposed_actions[None, :], random_proposed_actions], axis=0)
            else:
                all_proposed_actions = random_proposed_actions
        else:
            all_proposed_actions = proposed_actions[None, :]

        proposed_actions, loss = optimize_actions_multistart(
            loss_function=self.loss_function,
            grad_loss_function=self.grad_loss_function,
            all_proposed_actions=all_proposed_actions,
            model=model,
            optimizer=self.excitation_optimizer,
            init_obs=obs,
            init_state=state,
            density_estimate=density_estimate,
            n_opt_steps=self.n_opt_steps,
            tau=self.tau,
            consider_action_distribution=self.consider_action_distribution,
            target_distribution=self.target_distribution,
            penalty_function=self.penalty_function,
        )

        action, next_proposed_actions, density_estimate = self.process_propositions(
            obs=obs,
            density_estimate=density_estimate,
            proposed_actions=proposed_actions,
            expl_action_key=expl_action_key,
        )

        return action, next_proposed_actions, density_estimate, loss, expl_key

    @eqx.filter_jit
    def process_propositions(
        self,
        obs: jax.Array,
        density_estimate: DensityEstimate,
        proposed_actions: jax.Array,
        expl_action_key: jax.random.PRNGKey,
    ) -> tuple[jax.Array, jax.Array, DensityEstimate]:
        """Updates the proposed actions and density estimate for the next iteration after the
        optimization has been performed.

        Args:
            obs (jax.Array): The momentary observation at time step k
            density_estimate (DensityEstimate): The momentary estimate of the data density without
                the momentary step k
            proposed_actions (jax.Array): The (potentially optimized) proposition of actions to take
            expl_action_key (jax.random.PRNGKey): A random key for the exploration action

        Returns:
            action (jax.Array): The chosen action
            next_proposed_actions (jax.Array): An initial proposition for future actions
            density_estimate (DensityEstimate): The updated density estimate now incorporating
                the current step k
        """
        action = proposed_actions[0, :]

        if self.clip_action:
            action = jnp.clip(action, -1, 1)

        next_proposed_actions = proposed_actions.at[:-1, :].set(proposed_actions[1:, :])

        new_proposed_action = jax.random.uniform(key=expl_action_key, minval=-1, maxval=1)
        next_proposed_actions = next_proposed_actions.at[-1, :].set(new_proposed_action)

        # update grid KDE with y_k and u_k
        if self.consider_action_distribution:
            density_estimate = update_density_estimate_single_observation(
                density_estimate, jnp.concatenate([obs, action], axis=-1)
            )
        else:
            density_estimate = update_density_estimate_single_observation(density_estimate, obs)

        return action, next_proposed_actions, density_estimate



import jax
import argparse
import eval_dmpe
import eval_goats


rpms = [0, 3000, 5000, 7000, 9000]

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Run DMPE on the PMSM environment.")
    parser.add_argument("--DMPE", action=argparse.BooleanOptionalAction)
    parser.add_argument("--iGOATS", action=argparse.BooleanOptionalAction)

    args = parser.parse_args()

    if args.DMPE:
        model_names = ["RLS", "NODE", "PM"]
        consider_actions = [True]

        for consider_action in consider_actions:
            for model_name in model_names:
                for rpm in rpms:
                    eval_dmpe.main(rpm, model_name, consider_action)

    elif args.iGOATS:
        jax.config.update("jax_platform_name", "cpu")

        consider_action = True
        for rpm in rpms:
            eval_goats.main(rpm, consider_action)

    else:
        raise ValueError("Please specify either DMPE or iGOATS")



from typing import Callable
from tqdm import tqdm

import jax
import jax.numpy as jnp
import equinox as eqx
import optax

import exciting_environments as excenvs
from dmpe.algorithms.algorithm_utils import (
    consult_exciter,
    interact_and_observe,
    default_dmpe_parameterization,
)
from dmpe.excitation.excitation_utils import loss_function, Exciter
from dmpe.models.model_training import ModelTrainer
from dmpe.utils.density_estimation import (
    DensityEstimate,
    build_grid,
)
from dmpe.utils.metrics import JSDLoss


def excite_and_fit(
    n_time_steps: int,
    env: excenvs.CoreEnvironment,
    model: eqx.Module,
    obs: jax.Array,
    state: excenvs.CoreEnvironment.State,
    proposed_actions: jax.Array,
    exciter: Exciter,
    model_trainer: ModelTrainer,
    density_estimate: DensityEstimate,
    observations: jax.Array,
    actions: jax.Array,
    opt_state_model: optax.OptState,
    loader_key: jax.random.PRNGKey,
    expl_key: jax.random.PRNGKey,
    callback_every: int,
    callback: Callable | None = None,
) -> tuple[jax.Array, jax.Array, eqx.Module, DensityEstimate]:
    """
    Main algorithm to apply to a given (unknown) system and generate informative data from that system.

    A pseudocode description of this algorithm is given in the corresponding publication [Vater2024].
    In summary, the algorithm iterates over the time steps k. Each iteration an action is chosen to be
    applied by the Exciter object. Afterwards, the action is applied to the system and its effect is
    observed. This is followed up by the (optional) update to the dynamics model.
    The rest of the code is only for monitoring and, finally, progressing to the next time step k+1.

    Args:
        n_time_steps (int): The number of time steps to run the algorithm for.
        env (excenvs.CoreEnvironment): The environment object representing the system.
        model (eqx.Module): The model used for prediction.
        obs (jax.Array): The initial observation of the system.
        state (excenvs.CoreEnvironment.State): The initial state of the system.
        proposed_actions (jax.Array): The proposed actions for exploration.
        exciter (Exciter): The exciter object responsible for choosing actions.
        model_trainer (ModelTrainer): The model trainer object responsible for training the model.
        density_estimate (DensityEstimate): The density estimate used for exploration.
        observations (jax.Array): The history of observations.
        actions (jax.Array): The history of actions.
        opt_state_model (optax.OptState): The optimizer state for the model.
        loader_key (jax.random.PRNGKey): The key used for loading data.
        expl_key: (jax.random.PRNGKey): The key used for random action generation.
        callback_every (int): The frequency at which to run the callback function.
        callback (Callable | None): Implementation of the callback function.

    Returns:
        tuple[jax.Array, jax.Array, eqx.Module, DensityEstimate, list, jax.Array, list]: A tuple containing
        the history of observations, the history of actions, the updated model, the updated density estimate,
        the prediction losses, the proposed actions, and the callback output.
    """
    prediction_losses = []
    data_losses = []

    callback_out = []

    for k in tqdm(range(n_time_steps)):
        action, next_proposed_actions, next_density_estimate, prediction_loss, next_expl_key = consult_exciter(
            k=k,
            exciter=exciter,
            obs=obs,
            state=state,
            model=model,
            density_estimate=density_estimate,
            proposed_actions=proposed_actions,
            expl_key=expl_key,
        )

        prediction_losses.append(prediction_loss)  # predicted loss for the last excitation optimization

        next_obs, next_state, actions, observations = interact_and_observe(
            env=env, k=jnp.array([k]), action=action, state=state, actions=actions, observations=observations
        )

        if model_trainer is not None:
            if k > model_trainer.start_learning:
                model, opt_state_model, loader_key = model_trainer.fit(
                    model=model,
                    k=jnp.array([k]),
                    observations=observations,
                    actions=actions,
                    opt_state=opt_state_model,
                    loader_key=loader_key,
                )
        elif hasattr(model, "fit"):
            model = model.fit(model, jnp.array([k]), observations, actions)
        else:
            if k == 0:
                print("Model is used statically and not re-fitted or updated otherwise.")

        ## Start Monitoring
        # evaluate the current excitation metric value for the acquired data
        # (Only necessary for monitoring purposes)
        data_loss = JSDLoss(
            next_density_estimate.p / jnp.sum(next_density_estimate.p),
            exciter.target_distribution / jnp.sum(exciter.target_distribution),
        )
        data_losses.append(data_loss)

        # callback
        if k % callback_every == 0 and k > 0:
            if callback is not None:
                callback_out.append(
                    callback(
                        k=jnp.array([k]),
                        env=env,
                        obs=obs,
                        state=state,
                        action=action,
                        next_obs=next_obs,
                        next_state=next_state,
                        observations=observations,
                        actions=actions,
                        model=model,
                        density_estimate=density_estimate,
                        proposed_actions=proposed_actions,
                        next_density_estimate=next_density_estimate,
                        next_proposed_actions=next_proposed_actions,
                        data_losses=data_losses,
                        prediction_losses=prediction_losses,
                    )
                )
        ## End Monitoring

        # k <- k + 1
        obs = next_obs
        state = next_state
        proposed_actions = next_proposed_actions
        density_estimate = next_density_estimate
        expl_key = next_expl_key

    return observations, actions, model, density_estimate, prediction_losses, proposed_actions, callback_out


def excite_with_dmpe(
    env: excenvs.CoreEnvironment,
    exp_params: dict,
    proposed_actions: jax.Array,
    loader_key: jax.random.PRNGKey,
    expl_key: jax.random.PRNGKey,
    callback_every: int | None = None,
    callback: Callable | None = None,
):
    """
    Excite the system using the Differentiable Model Predictive Excitation (DMPE) algorithm.

    Args:
        env (excenvs.CoreEnvironment): The environment object representing the system.
        exp_params (dict): The experiment parameters.
        proposed_actions (jax.Array): The initial proposed actions to apply.
        loader_key (jax.random.PRNGKey): The key used for loading data.
        expl_key (jax.random.PRNGKey): The key used for random action generation.
        callback_every (int | None): The frequency of steps at which to run the callback function.
            If it is 'None' no callback is done.
        callback (Callable): Callback function to monitor the excitation process.
            See 'dmpe/evaluation/callbacks.py' for examples and API.

    Returns:
        Tuple[jax.Array, jax.Array, eqx.Module, DensityEstimate, list, jax.Array, list]: A tuple containing
        the history of observations, the history of actions, the updated model, the updated density estimate,
        the prediction losses, the proposed actions, and the callback output.
    """
    obs, state = env.reset(env.env_properties)

    dim_obs_space = obs.shape[0]
    dim_action_space = env.action_dim

    if exp_params["alg_params"]["consider_action_distribution"]:
        dim = dim_obs_space + dim_action_space
    else:
        dim = dim_obs_space

    n_grid_points = exp_params["alg_params"]["points_per_dim"] ** dim

    # setup memory variables
    observations = jnp.zeros((exp_params["n_time_steps"], dim_obs_space))
    observations = observations.at[0].set(obs)
    actions = jnp.zeros((exp_params["n_time_steps"] - 1, dim_action_space))

    exciter = Exciter(
        start_optimizing=exp_params["alg_params"]["start_optimizing"],
        loss_function=loss_function,
        grad_loss_function=jax.value_and_grad(loss_function, argnums=(3)),
        excitation_optimizer=exp_params["alg_params"]["excitation_optimizer"],
        tau=env.tau,
        n_opt_steps=exp_params["alg_params"]["n_opt_steps"],
        consider_action_distribution=exp_params["alg_params"]["consider_action_distribution"],
        target_distribution=exp_params["alg_params"]["target_distribution"],
        penalty_function=exp_params["alg_params"]["penalty_function"],
        clip_action=exp_params["alg_params"]["clip_action"],
        n_starts=exp_params["alg_params"]["n_starts"],
        reuse_proposed_actions=exp_params["alg_params"]["reuse_proposed_actions"],
    )

    if exp_params["model_trainer_params"] is None and exp_params["model_params"] is None:
        model_trainer = None
        model = env
        opt_state_model = None
    elif exp_params["model_trainer_params"] is None and exp_params["model_params"] is not None:
        model_trainer = None
        model = exp_params["model_class"](**exp_params["model_params"])
        opt_state_model = None
    else:
        model_trainer = ModelTrainer(
            start_learning=exp_params["model_trainer_params"]["start_learning"],
            training_batch_size=exp_params["model_trainer_params"]["training_batch_size"],
            n_train_steps=exp_params["model_trainer_params"]["n_train_steps"],
            sequence_length=exp_params["model_trainer_params"]["sequence_length"],
            featurize=exp_params["model_trainer_params"]["featurize"],
            model_optimizer=optax.adabelief(exp_params["model_trainer_params"]["model_lr"]),
            tau=env.tau,
        )
        model = exp_params["model_class"](**exp_params["model_params"])
        opt_state_model = model_trainer.model_optimizer.init(eqx.filter(model, eqx.is_inexact_array))

    density_estimate = DensityEstimate(
        p=jnp.zeros([n_grid_points, 1]),
        z_g=build_grid(
            dim,
            low=-exp_params["alg_params"]["grid_extend"],
            high=exp_params["alg_params"]["grid_extend"],
            points_per_dim=exp_params["alg_params"]["points_per_dim"],
        ),
        bandwidth=jnp.array([exp_params["alg_params"]["bandwidth"]]),
        n_observations=jnp.array([0]),
    )

    observations, actions, model, density_estimate, losses, proposed_actions, callback_out = excite_and_fit(
        n_time_steps=exp_params["n_time_steps"],
        env=env,
        model=model,
        obs=obs,
        state=state,
        proposed_actions=proposed_actions,
        exciter=exciter,
        model_trainer=model_trainer,
        density_estimate=density_estimate,
        observations=observations,
        actions=actions,
        opt_state_model=opt_state_model,
        loader_key=loader_key,
        expl_key=expl_key,
        callback_every=callback_every if callback_every is not None else exp_params["n_time_steps"] + 1,
        callback=callback,
    )

    return observations, actions, model, density_estimate, losses, proposed_actions, callback_out


def default_dmpe(env, seed=0, n_time_steps=5000, featurize=None, model_class=None, callback=None, callback_every=None):
    """Runs DMPE with default parameterization. The parameter choices might
    not be optimal for a given system.

    In future work, automatic tuning for the parameters will be added such that no
    manual tuning is required.

    Args:
        env: The environment object representing the system.
        seed (int): The random seed for reproducibility.
        n_time_steps (int): The number of time steps to run the algorithm for.
        featurize: The function used for feature applied onto the observations.
        model_class: The class of the model used for prediction.
        callback (Callable): Callback function to monitor the excitation process.
            See 'dmpe/evaluation/callbacks.py' for examples and API.
        callback_every (int | None): The frequency of steps at which to run the callback function.
            If it is 'None' no callback is done.

    Returns:
        Tuple[jax.Array, jax.Array, eqx.Module, DensityEstimate, list, jax.Array, list]: A tuple containing
        the history of observations, the history of actions, the updated model, the updated density estimate,
        the prediction losses, the proposed actions, and the callback output.
    """

    return excite_with_dmpe(
        env,
        *default_dmpe_parameterization(env, seed, n_time_steps, featurize, model_class),
        callback_every=callback_every,
        callback=callback,
    )



"""Smaller utils specifically for the experiments with the pmsm environment."""

import matplotlib.pyplot as plt
import numpy as np

import jax
import jax.numpy as jnp

from exciting_environments.pmsm.pmsm_env import PMSM
from dmpe.excitation.excitation_utils import soft_penalty
from dmpe.evaluation.plotting_utils import plot_sequence


class ExcitingPMSM(PMSM):

    def __init__(self, initial_rpm, *args, **kwargs):
        self.initial_rpm = initial_rpm
        super().__init__(*args, **kwargs)

    def generate_observation(self, system_state, env_properties):
        physical_normalizations = env_properties.physical_normalizations

        obs = jnp.hstack(
            (
                physical_normalizations.i_d.normalize(system_state.physical_state.i_d),
                physical_normalizations.i_q.normalize(system_state.physical_state.i_q),
            )
        )
        return obs

    def init_state(self, env_properties, rng=None, vmap_helper=None):
        """Returns default initial state for all batches."""
        phys = self.PhysicalState(
            u_d_buffer=0.0,
            u_q_buffer=0.0,
            epsilon=0.0,
            i_d=-env_properties.physical_normalizations.i_d.max / 2,
            i_q=0.0,
            torque=0.0,
            omega_el=2 * jnp.pi * 3 * self.initial_rpm / 60,
        )
        subkey = jnp.nan
        additions = None
        ref = self.PhysicalState(
            u_d_buffer=jnp.nan,
            u_q_buffer=jnp.nan,
            epsilon=jnp.nan,
            i_d=jnp.nan,
            i_q=jnp.nan,
            torque=jnp.nan,
            omega_el=jnp.nan,
        )
        return self.State(physical_state=phys, PRNGKey=subkey, additions=additions, reference=ref)


def PMSM_penalty(env, observations, actions, penalty_order=2):

    # action_penalty = soft_penalty(actions, a_max=0.8660254, penalty_order=penalty_order)
    # action_penalty = soft_penalty(actions, a_max=1, penalty_order=penalty_order)

    if actions is None:
        action_penalty = 0
    else:
        physical_u_d = env.env_properties.action_normalizations.u_d.denormalize(actions[..., 0]) / 400
        physical_u_q = env.env_properties.action_normalizations.u_q.denormalize(actions[..., 1]) / 400

        action_penalty = jax.nn.relu(jnp.sqrt(physical_u_d**2 + physical_u_q**2) - 1 / jnp.sqrt(3)) ** penalty_order
        action_penalty = jnp.sum(action_penalty)

    physical_i_d = env.env_properties.physical_normalizations.i_d.denormalize(observations[..., 0])
    physical_i_q = env.env_properties.physical_normalizations.i_q.denormalize(observations[..., 1])

    a = physical_i_d / jnp.abs(env.env_properties.physical_normalizations.i_d.min)
    b = physical_i_q / jnp.abs(env.env_properties.physical_normalizations.i_q.max)

    obs_penalty = jax.nn.relu(jnp.sqrt(a**2 + b**2) - 1) ** penalty_order
    obs_penalty = jnp.sum(obs_penalty)
    i_d_penalty = jnp.sum(jax.nn.relu(a)) ** penalty_order

    return (obs_penalty + i_d_penalty + action_penalty) * 1e3
    # return (obs_penalty + action_penalty) * 1e3


def plot_current_constraints(fig, ax, i_d_normalizer, i_q_normalizer):
    i_d_plot = np.linspace(i_d_normalizer.min, i_d_normalizer.max, 1000)
    i_q_plot = np.sqrt(i_q_normalizer.max**2 - i_d_plot**2)

    ax.plot(i_d_plot, i_q_plot, "k")
    ax.plot(i_d_plot, -i_q_plot, "k")
    ax.plot(np.zeros(2), np.array([i_q_normalizer.min, i_q_normalizer.max]), "k")
    return fig, ax


def plot_sequence_with_constraints(env, observations, actions):
    i_d_normalizer = env.env_properties.physical_normalizations.i_d
    i_q_normalizer = env.env_properties.physical_normalizations.i_q

    physical_i_d = i_d_normalizer.denormalize(observations[..., 0])
    physical_i_q = i_q_normalizer.denormalize(observations[..., 1])

    print("max i_q current:", env.env_properties.physical_normalizations.i_q.max, "A")
    print("min i_q current:", env.env_properties.physical_normalizations.i_q.min, "A")

    fig, axs = plot_sequence(
        observations=jnp.vstack([physical_i_d, physical_i_q]).T,
        actions=jnp.vstack(
            [
                env.env_properties.action_normalizations.u_d.denormalize(actions[..., 0]),
                env.env_properties.action_normalizations.u_q.denormalize(actions[..., 1]),
            ]
        ).T,
        tau=env.tau,
        obs_labels=["i_d", "i_q"],
        action_labels=["u_d", "u_q"],
    )
    t = jnp.linspace(0, observations.shape[0] - 1, observations.shape[0]) * env.tau
    axs[0].plot(t, env.env_properties.physical_normalizations.i_d.denormalize(np.ones(observations.shape[0])), "k")
    axs[0].plot(t, env.env_properties.physical_normalizations.i_d.denormalize(-np.ones(observations.shape[0])), "k")
    axs[0].plot(t, env.env_properties.physical_normalizations.i_q.denormalize(np.ones(observations.shape[0])), "k")
    axs[0].set_ylim(-350, 350)

    axs[1].set_xlim(-350, 100)
    axs[1].set_ylim(-350, 350)
    axs[2].plot(t[:-1], env.env_properties.action_normalizations.u_d.denormalize(np.ones(actions.shape[0])), "k")
    axs[2].plot(t[:-1], env.env_properties.action_normalizations.u_d.denormalize(-np.ones(actions.shape[0])), "k")

    plot_current_constraints(fig, axs[1], i_d_normalizer, i_q_normalizer)

    return fig, axs


def plot_constraints_induced_voltage(env, physical_i_d, physical_i_q, w_el, saturated=True, show_torque=False):
    """Plot the constraints onto the currents due to the induced voltage."""
    p = env.env_properties.static_params.p
    r_s = env.env_properties.static_params.r_s
    psi_p = env.env_properties.static_params.psi_p

    if saturated:
        Psi_d = env.pmsm_lut["Psi_d"][1:-1, 1:-1]
        Psi_q = env.pmsm_lut["Psi_q"][1:-1, 1:-1]
        i_d_vec = np.squeeze(env.pmsm_lut["i_d_vec"])
        i_q_vec = np.squeeze(env.pmsm_lut["i_q_vec"])

    if not saturated:
        # initialize flux matrices and current vectors or something

        l_d = env.env_properties.static_params.l_d
        l_q = env.env_properties.static_params.l_q

    udc = 400  # This should really be part of the properties somehow

    U = np.zeros_like(Psi_d)
    I = np.zeros_like(Psi_d)
    T = np.zeros_like(Psi_d)
    Id = np.zeros_like(Psi_d)
    Iq = np.zeros_like(Psi_d)

    # Loop over current vectors
    w = w_el
    for dd in range(len(i_d_vec)):
        for qq in range(len(i_q_vec)):
            id_val = float(i_d_vec[dd])
            iq_val = float(i_q_vec[qq])

            if saturated:
                psid = Psi_d[qq, dd]
                psiq = Psi_q[qq, dd]
            else:
                psid = psi_p + l_d * id_val
                psiq = l_q * iq_val

            ud = r_s * id_val - w * psiq
            uq = r_s * iq_val + w * psid

            Id[qq, dd] = id_val
            Iq[qq, dd] = iq_val
            U[qq, dd] = np.sqrt(ud**2 + uq**2)
            I[qq, dd] = np.sqrt(id_val**2 + iq_val**2)
            T[qq, dd] = 1.5 * p * (psid * iq_val - psiq * id_val)

    Psi = np.sqrt(Psi_d**2 + Psi_q**2)

    # Plot results
    fig, ax = plt.subplots()
    ax.grid(True)
    ax.set_xlabel("id")
    ax.set_ylabel("iq")

    ax.scatter(jnp.squeeze(physical_i_d), jnp.squeeze(physical_i_q), s=1)

    umult = 1 / jnp.sqrt(3)  # 2 / 3  # * jnp.sqrt(2)
    ax.contour(Id, Iq, I, levels=[250], colors="k", linewidths=1.5)
    if show_torque:
        ax.contour(Id, Iq, T, linestyles="dashed", colors="k")
    ax.contour(Id, Iq, U, levels=[udc * umult], colors="k", linewidths=1.5)

    ax.set_ylim(-350, 350)
    ax.set_xlim((-350, 200))

    return fig, ax



import numpy as np
import matplotlib.pyplot as plt

import jax
import jax.numpy as jnp

from dmpe.models.model_utils import simulate_ahead, simulate_ahead_with_env
from dmpe.models.model_training import precompute_starting_points, load_single_batch
from dmpe.evaluation.plotting_utils import plot_sequence_and_prediction, plot_model_performance


def callback_template(**kwargs):
    """Template callback function to show what keyword arguments are available.

    Args:
        **kwargs: Keyword arguments passed to the callback.
    """
    for kwargs_key, kwargs_value in kwargs.items():
        print(f"{kwargs_key}: {type(kwargs_value)}")


def plot_sequence_callback(**kwargs):
    """Callback that plots the sequence of observations and actions and the current
    prediction of the model that will be used to initialize the optimization of the
    next action.
    """
    k = kwargs["k"].item()
    fig, axs = plot_sequence_and_prediction(
        observations=kwargs["observations"][: k + 2, :],
        actions=kwargs["actions"][: k + 1, :],
        tau=kwargs["env"].tau,
        obs_labels=kwargs["env"].obs_description,
        actions_labels=kwargs["env"].action_description,
        model=kwargs["model"],
        init_obs=kwargs["next_obs"],
        init_state=kwargs["next_state"],
        proposed_actions=kwargs["next_proposed_actions"],
    )

    plt.show()


def model_performance_callback(sequence_length: int = 10, batch_size: int = 5, show_plot: bool = True, **kwargs):
    """Callback that evaluates the model performance on a batch of data.

    Args:
        sequence_length (int): Length of the sequence to evaluate.
        batch_size (int): Number of sequences to evaluate.
        show_plot (bool): Whether to show the plot of the model performance.
        **kwargs: Keyword arguments passed to the callback.

    Returns:
        mses (float): Mean squared error of the model predictions averaged over the
            trajectories and the batch of trajectories.
    """
    starting_points, _ = precompute_starting_points(
        n_train_steps=1,
        k=kwargs["k"],
        sequence_length=sequence_length,
        training_batch_size=batch_size,
        loader_key=jax.random.key(seed=np.random.randint(low=100_000)),
    )

    batched_obs, batched_act = load_single_batch(
        observations_array=kwargs["observations"],
        actions_array=kwargs["actions"],
        starting_points=starting_points[0, ...],
        sequence_length=sequence_length,
    )

    mses = []

    model = kwargs["model"]
    env = kwargs["env"]

    for i in range(batched_obs.shape[0]):
        if show_plot:
            fig, _, pred_observations = plot_model_performance(
                model,
                batched_obs[i, ...],
                batched_act[i, ...],
                env.tau,
                env.obs_description,
                env.action_description,
            )
            plt.show()
        else:
            pred_observations = simulate_ahead(
                model,
                batched_obs[i, 0, :],
                batched_act[i, ...],
                env.tau,
            )

        mse = jnp.mean((pred_observations - batched_obs[i, ...]) ** 2)
        mses.append(mse)

    return np.mean(mses)


def optimization_prediction_accuracy_callback(**kwargs):
    """Callback that evaluates the accuracy of the prediction accuracy for the momentary
    iteration, i.e., explicitly for the actions that have been chosen based on the internal
    optimization loop. This is done by comparing the predicted observations with the actual
    observations and calculating the mean squared error.

    Currently, this callback is only implemented for two observation dimensions.
    """
    env = kwargs["env"]
    model = kwargs["model"]

    sim_actions = jnp.concatenate([kwargs["action"][None], kwargs["next_proposed_actions"][:-1]], axis=0)
    model_prediction = simulate_ahead(model, kwargs["obs"], sim_actions, env.tau)
    env_prediction, _ = simulate_ahead_with_env(env, kwargs["obs"], kwargs["state"], sim_actions)

    fig, axs = plt.subplots(nrows=1, ncols=3, figsize=(25, 6))

    axs[0].plot(model_prediction[..., 0], label="model")
    axs[0].plot(env_prediction[..., 0], label="env")

    axs[1].plot(model_prediction[..., 1], label="model")
    axs[1].plot(env_prediction[..., 1], label="env")

    axs[2].plot((model_prediction[..., 0] - env_prediction[..., 0]), "r", label=env.obs_description[0])
    axs[2].plot((model_prediction[..., 1] - env_prediction[..., 1]), "k", label=env.obs_description[1])

    for ax in axs:
        ax.grid(True)
        ax.legend()

    fig.suptitle("Optimization result accuracy")
    plt.show()



from typing import Tuple, Callable

import jax
import jax.numpy as jnp
import equinox as eqx
import optax
from haiku import PRNGSequence

import exciting_environments as excenvs
from dmpe.utils.signals import aprbs
from dmpe.utils.density_estimation import DensityEstimate, get_uniform_target_distribution
from dmpe.models.models import NeuralEulerODE
from dmpe.excitation.excitation_utils import soft_penalty, Exciter


def consult_exciter(
    k: int,
    exciter: Exciter,
    obs: jax.Array,
    state: excenvs.CoreEnvironment.State,
    model: eqx.Module,
    density_estimate: DensityEstimate,
    proposed_actions: jax.Array,
    expl_key: jax.random.PRNGKey,
):
    """Use the exciter to choose the next action or simply apply the proposed actions depending on
    the current time step.

    This also updates the density estimate based on the chosen action. The momentary observation
    y_k has no corresponding action u_k, before the exciter has chosen one. Therefore, the
    DensityEstimate that goes into the function is still based on y_{k-1} and u_{k-1}. The
    one that is returned from this function incorporates y_k and u_k.

    This is essentially a wrapper around the exciter's choose_action method that enables
    overwriting the function for certain cases. Especially, this is used here to allow for random
    actions at the start of the experiment where the model is not yet trained and random actions
    are more effective compared to actions optimized based on the poor model prediction.

    Args:
        k (int): The current time step.
        exciter (Exciter): The exciter object.
        obs(jax.Array): The current observation.
        state(excenvs.CoreEnvironment.State): The current state of the environment.
        model (eqx.Module): The model used for predictions.
        density_estimate (DensityEstimate): The density estimate object from time step k-1.
        proposed_actions (jax.Array): The proposed actions for the current time step.
        expl_key (jax.random.PRNGKey): The random key for drawing new proposed actions.

    Returns:
        action (jax.Array): The chosen action.
        next_proposed_actions (jax.Array): The proposed actions for the next time step.
        next_density_estimate (DensityEstimate): The updated density estimate with (y_k, u_k).
        prediction_loss (float): The prediction loss for the current time step.
        next_expl_key (jax.random.PRNGKey): The updated random key for drawing new proposed actions.
    """
    if k > exciter.start_optimizing:
        action, next_proposed_actions, next_density_estimate, prediction_loss, next_expl_key = exciter.choose_action(
            obs=obs,
            state=state,
            model=model,
            density_estimate=density_estimate,
            proposed_actions=proposed_actions,
            expl_key=expl_key,
        )
    else:
        # run the exciter without optimizing actions
        next_expl_key, expl_action_key = jax.random.split(expl_key, 2)
        action, next_proposed_actions, next_density_estimate = exciter.process_propositions(
            obs=obs,
            density_estimate=density_estimate,
            proposed_actions=proposed_actions,
            expl_action_key=expl_action_key,
        )
        prediction_loss = 0.0  # no prediction loss since no optimization has been performed

    return action, next_proposed_actions, next_density_estimate, prediction_loss, next_expl_key


@eqx.filter_jit
def interact_and_observe(
    env: excenvs.CoreEnvironment,
    k: jax.Array,
    action: jax.Array,
    state: excenvs.CoreEnvironment.State,
    actions: jax.Array,
    observations: jax.Array,
) -> Tuple[jax.Array, excenvs.CoreEnvironment.State, jax.Array, jax.Array]:
    """
    Interact with the environment and store the action and the resulting observation.

    Args:
        env (excenvs.CoreEnvironment): The environment object.
        k (int): The current time step.
        action (jax.Array): The action to be taken at time step k.
        state (excenvs.CoreEnvironment.State): The state of the environment at time step k.
        actions (jax.Array): The array of actions taken so far.
        observations (jax.Array): The array of observations observed so far.

    Returns:
        obs (jax.Array): The updated observation at time step k+1.
        state (excenvs.CoreEnvironment.State): The updated state of the environment at time step k+1.
        actions (jax.Array): The updated array of actions taken so far.
        observations (jax.Array): The updated array of observations observed so far.
    """

    # apply u_k and go to x_{k+1} and observe y_{k+1}
    obs, state = env.step(state, action, env.env_properties)

    actions = actions.at[k].set(action)  # store u_k
    observations = observations.at[k + 1].set(obs)  # store y_{k+1}

    return obs, state, actions, observations


def default_dmpe_parameterization(
    env: excenvs.CoreEnvironment,
    seed: int = 0,
    n_time_steps=5_000,
    featurize: Callable | None = None,
    model_class: eqx.Module | None = None,
):
    """Returns a default parameterization for the DMPE algorithm.

    This parameterization is intended as a starting point to apply to a given system.
    The parameters are not necessarily optimal for any given system but should give a
    reasonable first impression. Currently, featurization of the model state e.g. angles
    needs to be provided manually.

    Args:
        env (excenvs.CoreEnvironment): The environment object representing the system.
        seed (int): The seed for the random number generator.
        n_time_steps (int): The number of time steps for the experiment.
        featurize (callable | None): A function to featurize the model state. Defaults to the identity function.
        model_class (eqx.Module | None): The model class to be used. Defaults to NeuralEulerODE. It must comply
            with the NeuralEulerODE API to be usable here. Otherwise, the model can also be overwritten after
            getting the other default parameters.

    Returns:
        Tuple[Dict, jax.Array, jax.random.PRNGKey, jax.random.PRNGKey]: A tuple containing the experiment parameters,
        the initial proposed actions, the key for loading data in model learning, and the key for random action generation.

    """
    alg_params = dict(
        bandwidth=0.08,
        n_prediction_steps=10,
        points_per_dim=21,
        grid_extend=1.05,
        excitation_optimizer=optax.adabelief(1e-2),
        n_opt_steps=50,
        start_optimizing=5,
        consider_action_distribution=True,
        penalty_function=None,
        target_distribution=None,
        clip_action=False,
        n_starts=10,
        reuse_proposed_actions=True,
    )

    dim = env.physical_state_dim + env.action_dim

    alg_params["penalty_function"] = lambda x, u: soft_penalty(a=x, a_max=1, penalty_order=2) + soft_penalty(
        a=u, a_max=1, penalty_order=2
    )
    alg_params["target_distribution"] = get_uniform_target_distribution(
        dim=3 if alg_params["consider_action_distribution"] else 2,
        points_per_dim=alg_params["points_per_dim"],
        bandwidth=alg_params["bandwidth"],
        grid_extend=alg_params["grid_extend"],
        consider_action_distribution=alg_params["consider_action_distribution"],
        penalty_function=alg_params["penalty_function"],
    )

    model_trainer_params = dict(
        start_learning=alg_params["n_prediction_steps"],
        training_batch_size=128,
        n_train_steps=1,
        sequence_length=alg_params["n_prediction_steps"],
        featurize=(lambda x: x) if featurize is None else featurize,
        model_lr=1e-4,
    )

    model_params = dict(
        obs_dim=env.reset(env.env_properties)[0].shape[0], action_dim=env.action_dim, width_size=128, depth=3, key=None
    )

    exp_params = dict(
        seed=seed,
        n_time_steps=n_time_steps,
        model_class=NeuralEulerODE if model_class is None else model_class,
        env_params=None,
        alg_params=alg_params,
        model_trainer_params=model_trainer_params,
        model_params=model_params,
    )

    key = jax.random.PRNGKey(seed=exp_params["seed"])
    data_key, model_key, loader_key, expl_key, key = jax.random.split(key, 5)

    data_rng = PRNGSequence(data_key)
    exp_params["model_params"]["key"] = model_key

    # initial guess
    proposed_actions = jnp.hstack(
        [
            aprbs(alg_params["n_prediction_steps"], env.batch_size, 1, 10, next(data_rng))[0]
            for _ in range(env.action_dim)
        ]
    )

    return exp_params, proposed_actions, loader_key, expl_key



import json
import argparse
import datetime
import os
import pathlib


import numpy as np
import jax
import jax.numpy as jnp
import diffrax
from haiku import PRNGSequence

os.environ["XLA_PYTHON_CLIENT_PREALLOCATE"] = "false"


from dmpe.utils.signals import aprbs
from dmpe.algorithms.algorithms import excite_with_dmpe
from dmpe.models.model_utils import save_model

import dmpe.utils.env_utils.pmsm_utils as pmsm_utils
import dmpe.evaluation.scripts.pmsm.dmpe_params as dmpe_params


TARGETED_DATA_PATH = (
    pathlib.Path(__file__).parent.parent.parent.parent.parent / pathlib.Path("data") / pathlib.Path("pmsm")
)


def safe_json_dump(obj, fp):
    default = lambda o: f"<<non-serializable: {type(o).__qualname__}>>"
    return json.dump(obj, fp, default=default)


def setup_env(rpm):
    env = pmsm_utils.ExcitingPMSM(
        initial_rpm=rpm,
        batch_size=1,
        saturated=True,
        LUT_motor_name="BRUSA",
        static_params={
            "p": 3,
            "r_s": 17.932e-3,
            "l_d": jnp.nan,
            "l_q": jnp.nan,
            "psi_p": 65.65e-3,
            "deadtime": 0,
            "u_dc": 400,
        },
        solver=diffrax.Tsit5(),
    )
    penalty_function = lambda observations, actions: pmsm_utils.PMSM_penalty(env, observations, actions)

    return env, penalty_function


def run_experiment(model_name, exp_idx, env, exp_params):

    seed = exp_params["seed"]
    rpm = exp_params["rpm"]
    consider_actions = exp_params["alg_params"]["consider_action_distribution"]

    print(
        "Running experiment",
        exp_idx,
        f"(seed: {int(seed)}) on the PMSM with {rpm} rpm. Considers actions? {consider_actions}",
    )

    # Check that the targeted data folder actually exist:
    results_path = TARGETED_DATA_PATH / pathlib.Path("dmpe") / pathlib.Path(model_name)
    print(f"Results will be written to: '{results_path}'.")
    assert results_path.exists(), (
        f"The expected results path '{results_path}' does not seem to exist. Please create the necessary file structure "
        + "or adapt the path."
    )

    # setup PRNG:
    key = jax.random.PRNGKey(seed=exp_params["seed"])
    data_key, model_key, loader_key, expl_key, key = jax.random.split(key, 5)
    data_rng = PRNGSequence(data_key)

    if model_name == "NODE":
        exp_params["model_params"]["key"] = model_key

    # initial guess
    proposed_actions = (
        jnp.hstack(
            [
                aprbs(exp_params["alg_params"]["n_prediction_steps"], env.batch_size, 1, 10, next(data_rng))[0]
                for _ in range(env.action_dim)
            ]
        )
        / 5
    )

    # run excitation algorithm
    observations, actions, model, density_estimate, losses, proposed_actions, _ = excite_with_dmpe(
        env, exp_params, proposed_actions, loader_key, expl_key, callback_every=None, callback=None
    )

    # save parameters
    file_name = datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
    with open(results_path / pathlib.Path(f"params_rpm_{rpm}_ca_{consider_actions}_{file_name}.json"), "w") as fp:
        safe_json_dump(exp_params, fp)

    # save observations + actions
    with open(results_path / pathlib.Path(f"data_rpm_{rpm}_ca_{consider_actions}_{file_name}.json"), "w") as fp:
        json.dump(dict(observations=observations.tolist(), actions=actions.tolist()), fp)

    model_params = exp_params["model_params"]

    if model_name == "NODE":
        model_params["key"] = model_params["key"].tolist()
    save_model(
        results_path / pathlib.Path(f"model_rpm_{rpm}_ca_{consider_actions}_{file_name}.json"),
        hyperparams=model_params,
        model=model,
    )


def main(rpm, model_name, consider_actions):

    assert 0 <= rpm <= 11000, "RPM must be between 0 and 11000."

    env, penalty_function = setup_env(rpm)

    if model_name == "NODE":
        alg_params, model_params, model_class, model_trainer_params = dmpe_params.get_NODE_params(
            consider_action_distribution=consider_actions, penalty_function=penalty_function
        )
    elif model_name == "RLS":
        alg_params, model_params, model_class, model_trainer_params = dmpe_params.get_RLS_params(
            consider_action_distribution=consider_actions, penalty_function=penalty_function
        )
    elif model_name == "PM":
        alg_params, model_params, model_class, model_trainer_params = dmpe_params.get_PM_params(
            consider_action_distribution=consider_actions, penalty_function=penalty_function
        )
    else:
        raise ValueError(f"Unknown model name: {model_name}")

    exp_params = dict(
        seed=None,
        rpm=float(rpm),
        n_time_steps=15_000,
        alg_params=alg_params,
        model_params=model_params,
        model_class=model_class,
        model_trainer_params=model_trainer_params,
    )
    seeds = list(np.arange(50, 61))

    for exp_idx, seed in enumerate(seeds):

        exp_params["seed"] = int(seed)
        run_experiment(model_name, exp_idx, env, exp_params)

        jax.clear_caches()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Run DMPE on the PMSM environment.")
    parser.add_argument("--rpm", type=float, default=2000, help="RPM of the PMSM.")
    parser.add_argument("--model", type=str, default="NODE", help="Model to use for the DMPE algorithm.")
    parser.add_argument("--consider_actions", action=argparse.BooleanOptionalAction)
    parser.add_argument("--gpu_id", type=int, default=0, help="GPU id to use.")

    args = parser.parse_args()

    gpus = jax.devices()
    jax.config.update("jax_default_device", gpus[args.gpu_id])

    rpm = args.rpm
    model_name = args.model
    consider_actions = args.consider_actions is True

    main(rpm, model_name, consider_actions)



from functools import partial

import jax
import jax.numpy as jnp
from exciting_environments.pmsm.pmsm_env import dq2albet, ROTATION_MAP


def get_advanced_angle(eps, tau_scale, tau, omega):
    return eps + tau_scale * tau * omega


class ClassicController:
    def __init__(self, motor, rpm, saturated, a=4, decoupling=True, tau=1e-4):
        """
        Initializes the ClassicController.

        Args:
            motor (Motor): The motor object containing motor parameters.
            saturated (bool): Indicates whether the motor is saturated.
            a (int, optional): A parameter for the controller. Defaults to 4.
            decoupling (bool, optional): Indicates whether decoupling is enabled. Defaults to True.
            tau (float, optional): The time constant for the controller. Defaults to 1e-4.

        Attributes:
            p_gain (float): Proportional gain for the controller.
            i_gain (float): Integral gain for the controller.
            tau (float): The time constant for the controller.
            a (int): A parameter for the controller.
            motor (Motor): The motor object containing motor parameters.
            L_dq (float): Inductance in the dq-axis (if not saturated).
            psi_dq (float): Flux linkage in the dq-axis (if not saturated).
            batch_size (int): The batch size for processing.
            decoupling (bool): Indicates whether decoupling is enabled.
            saturated (bool): Indicates whether the motor is saturated.
            interpolators (Interpolator): Interpolators for the motor (if saturated).
            u_s_0 (float): Initial control signal.
            integrated (float): Integrated error for the controller.
        """
        self.tau: float = tau
        self.a: int = a

        self.motor = motor
        self.rpm: float = rpm

        if not saturated:
            self.L_dq: float = motor.l_dq
            self.psi_dq: float = motor.psi_dq

        self.batch_size: int = None
        self.decoupling: bool = decoupling
        self.saturated: bool = saturated
        if self.saturated:
            self.interpolators = motor.LUT_interpolators
        self.u_s_0 = 0

        self.tune()

    def tune(self):
        """
        Tunes the controller by setting the proportional and integral gains.

        If the motor is not saturated, the proportional gain (p_gain) and integral gain (i_gain)
        are calculated based on the motor's inductance (L_dq), the parameter 'a', and the time constant (tau).

        Returns:
            None
        """
        if not self.saturated:
            self.p_gain = 1 / (self.a * 1.5 * self.tau) * self.L_dq
            self.i_gain = self.p_gain / ((self.a) ** 2 * 1.5 * self.tau)

    def check_constraints(self, e, eps, integrated, p_gain, i_gain, u_s_0):
        """
        Checks whether operation points are allowed to be integrated and returns a mask.

        The mask indicates whether the operation points are within the allowed constraints.
        A value of 0 indicates "nok" and 1 indicates "ok".

        Args:
            e (np.ndarray): The error array.
            eps (float): The epsilon value.

        Returns:
            np.ndarray: A mask array where 0s indicate not okay and 1s indicate okay.
        """

        mask = jnp.zeros_like(e)
        u_dq = p_gain * e + i_gain * (integrated + e * self.tau) + u_s_0

        u_dq_norm = self.motor.env_properties.action_normalizations.u_d.normalize(u_dq)

        u_albet_norm = dq2albet(
            u_dq_norm,
            get_advanced_angle(eps, 0.5, self.tau, (3 * self.rpm / 60 * 2 * jnp.pi)),
        )
        u_albet_c = u_albet_norm[:, 0] + 1j * u_albet_norm[:, 1]
        idx = (jnp.sin(jnp.angle(u_albet_c)[..., jnp.newaxis] - 2 / 3 * jnp.pi * jnp.arange(3)) >= 0).astype(int)
        rot_vecs = ROTATION_MAP[idx[:, 0], idx[:, 1], idx[:, 2]]
        u_albet_c = jnp.multiply(u_albet_c, rot_vecs)
        cond1 = (jnp.abs(u_albet_c.real) < 2 / 3) & (
            (u_albet_c.imag > 0) & (u_albet_c.imag < 2 / 3 * jnp.sqrt(3))
        )  # Inside
        cond2 = (
            ~(jnp.abs(u_albet_c.real) < 2 / 3) | ((u_albet_c.imag < 0) | (u_albet_c.imag > 2 / 3 * jnp.sqrt(3)))
        ) & jnp.all(
            e < 0, axis=1
        )  # outside but neg

        mask = jnp.logical_or(cond1, cond2)
        return mask

    def integrate(self, integrated, e, mask):
        """
        Integrates the error over time with the given mask.

        Args:
            e (np.ndarray): The error array.
            mask (np.ndarray): The mask array indicating which errors to integrate.

        Returns:
            None
        """
        # integrated += e * self.tau * mask
        integrated = (integrated + e * self.tau) * mask
        return integrated

    @partial(jax.jit, static_argnums=[0])
    def __call__(self, obs, integrated):
        """
        Computes the control action based on the observed state.

        Args:
            obs (np.ndarray): The observation array containing the normalized state, epsilon, and reference values.

        Returns:
            np.ndarray: The normalized control action voltages.
        """
        # Extract normalized state, epsilon, and reference values from the observation
        state_norm = obs[:, :2]
        eps = obs[:, 2:3]
        reference_norm = obs[:, 3:5]

        # Calculate the error between the reference and the current state
        state = jnp.concatenate(
            [
                self.motor.env_properties.physical_normalizations.i_d.denormalize(state_norm[:, 0]),
                self.motor.env_properties.physical_normalizations.i_q.denormalize(state_norm[:, 1]),
            ],
            axis=0,
        )[None]
        reference = jnp.concatenate(
            [
                self.motor.env_properties.physical_normalizations.i_d.denormalize(reference_norm[:, 0]),
                self.motor.env_properties.physical_normalizations.i_q.denormalize(reference_norm[:, 1]),
            ],
            axis=0,
        )[None]

        e = reference - state

        if self.saturated:
            # Tune gains if the motor is saturated
            p_d = {q: interp(state) for q, interp in self.interpolators.items()}
            # Neglect non-main-diagonal entries of L_diff
            L_dq = jnp.column_stack([p_d[q] for q in ["L_dd", "L_qq"]])
            psi_dq = jnp.column_stack([p_d[q] for q in ["Psi_d", "Psi_q"]])
            p_gain = 1 / (self.a * 1.5 * self.tau) * L_dq
            p_gain_help = 1 / (self.a * 1.5 * self.tau) * L_dq
            i_gain = 0.3 * p_gain_help / ((self.a) ** 2 * 1.5 * self.tau)

        if self.decoupling:
            # Apply decoupling if enabled
            i_dq = state
            q = jnp.array([0, -1, 1, 0]).reshape(2, 2)
            if self.saturated:
                # Calculate the initial control signal for saturated motor
                u_s_0 = (3 * self.rpm / 60 * 2 * jnp.pi) * jnp.einsum("ij,bj->bi", q, psi_dq)
            else:
                # Calculate the initial control signal for non-saturated motor
                u_s_0 = (3 * self.rpm / 60 * 2 * jnp.pi) * jnp.einsum("ij,bj->bi", q, i_dq * L_dq + psi_dq)
        else:
            u_s_0 = self.u_s_0

        # Check constraints and get the integration allowance mask
        integration_allowance_mask = self.check_constraints(e, eps, integrated, p_gain, i_gain, u_s_0)
        # Integrate the error with the mask
        integrated = self.integrate(integrated, e, integration_allowance_mask)
        # Calculate the control action
        u_dq = p_gain * e + i_gain * integrated + u_s_0

        # Normalize the control action voltages
        return self.motor.env_properties.action_normalizations.u_d.normalize(u_dq), integrated

    def reset(self, batch_size):
        """
        Resets the controller's state for a new batch of data.

        Args:
            batch_size (int): The size of the new batch.

        Returns:
            None
        """
        self.batch_size = batch_size

        return jnp.zeros((self.batch_size, 2))

    def parameters(self):
        """
        Returns the parameters of the controller.

        Returns:
            list: An empty list of parameters.
        """
        return []



import json
import argparse
import datetime
import os
import pathlib


import numpy as np
import jax

os.environ["CUDA_VISIBLE_DEVICES"] = "0"

jax.config.update("jax_platform_name", "cpu")


from dmpe.related_work.algorithms import excite_with_iGOATS
from dmpe.evaluation.scripts.pmsm.eval_dmpe import setup_env, safe_json_dump, TARGETED_DATA_PATH
from dmpe.evaluation.scripts.pmsm.goats_params import get_alg_params


def run_experiment(exp_idx, env, exp_params):

    seed = exp_params["seed"]
    rpm = exp_params["rpm"]
    consider_actions = exp_params["alg_params"]["consider_action_distribution"]

    print(
        "Running experiment",
        exp_idx,
        f"(seed: {int(seed)}) on the PMSM with {rpm} rpm. Considers actions? {consider_actions}",
    )

    # Check that the targeted data folder actually exist:
    results_path = TARGETED_DATA_PATH / pathlib.Path("igoats")
    print(f"Results will be written to: '{results_path}'.")
    assert results_path.exists(), (
        f"The expected results path '{results_path}' does not seem to exist. Please create the necessary file structure "
        + "or adapt the path."
    )

    # run excitation algorithm
    if not consider_actions:
        raise NotImplementedError("This code does not has no option to not take actions into account")

    alg_params = exp_params["alg_params"]
    observations, actions = excite_with_iGOATS(
        n_time_steps=exp_params["n_time_steps"],
        env=env,
        prediction_horizon=alg_params["prediction_horizon"],
        application_horizon=alg_params["application_horizon"],
        bounds_amplitude=alg_params["bounds_amplitude"],
        bounds_duration=alg_params["bounds_duration"],
        population_size=alg_params["population_size"],
        n_generations=alg_params["n_generations"],
        featurize=alg_params["featurize"],
        rng=np.random.default_rng(seed),
        compress_data=alg_params["compress_data"],
        compression_target_N=alg_params["compression_target_N"],
        compression_feat_dim=alg_params["compression_feat_dim"],
        compression_dist_th=alg_params["compression_dist_th"],
        penalty_function=alg_params["penalty_function"],
        plot_subsequences=False,
    )

    observations = [obs.tolist() for obs in observations]
    actions = [act.tolist() for act in actions]

    # save parameters
    file_name = datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
    with open(results_path / pathlib.Path(f"params_rpm_{rpm}_ca_{consider_actions}_{file_name}.json"), "w") as fp:
        safe_json_dump(exp_params, fp)

    # save observations + actions
    with open(results_path / pathlib.Path(f"data_rpm_{rpm}_ca_{consider_actions}_{file_name}.json"), "w") as fp:
        json.dump(dict(observations=observations, actions=actions), fp)


def main(rpm, consider_actions):

    assert 0 <= rpm <= 11000, "RPM must be between 0 and 11000."

    env, penalty_function = setup_env(rpm)

    alg_params = get_alg_params(consider_actions, penalty_function)

    exp_params = dict(
        seed=None,
        rpm=float(rpm),
        n_time_steps=15_000,
        alg_params=alg_params,
        model_params=None,
        model_class=None,
        model_trainer_params=None,
    )
    seeds = list(np.arange(50, 61))

    for exp_idx, seed in enumerate(seeds):

        exp_params["seed"] = int(seed)
        run_experiment(exp_idx, env, exp_params)

        jax.clear_caches()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Run DMPE on the PMSM environment.")
    parser.add_argument("--rpm", type=float, default=2000, help="RPM of the PMSM.")
    parser.add_argument("--consider_actions", action=argparse.BooleanOptionalAction)

    args = parser.parse_args()

    jax.config.update("jax_platform_name", "cpu")

    rpm = args.rpm
    consider_actions = args.consider_actions is True

    main(rpm, consider_actions)



from functools import partial

import jax
import jax.numpy as jnp


@partial(jax.jit, static_argnums=(0, 1))
def generate_constant_action(n_steps: int, batch_size: int, key: jax.random.PRNGKey) -> jax.Array:
    """Randomly draws an action and repeats it for 'n_steps'."""
    actions = jax.random.uniform(key, shape=(batch_size, 1, 1), minval=-1, maxval=1)
    actions = jnp.repeat(actions, repeats=n_steps, axis=1)
    return actions


def aprbs_single_batch(len: int, t_min: float, t_max: float, key: jax.random.PRNGKey) -> jax.Array:
    """Creates an amplitude modulated pseudorandom binary sequence (APRBS) in 1d and for 1 batch,
    i.e. without a batch dimension.

    Args:
        len (int): Length of the signal.
        t_min (float): Minimum hold time of an amplitude.
        t_max (float): Maximum hold time of an amplitude
        key (jax.random.PRNGKey): Random key for JAX random sampling.
    """

    t = 0
    sig = []
    while t < len:
        steps_key, value_key, key = jax.random.split(key, 3)

        t_step = jax.random.randint(steps_key, shape=(1,), minval=t_min, maxval=t_max)

        sig.append(jnp.ones(t_step) * jax.random.uniform(value_key, shape=(1,), minval=-1, maxval=1))
        t += t_step.item()

    return jnp.hstack(sig)[:len]


def aprbs(n_steps: int, batch_size: int, t_min: float, t_max: float, key: jax.random.PRNGKey) -> jax.Array:
    """Creates an amplitude modulated pseudorandom binary sequence (APRBS) in 1d where 'batch_size'
    sequences are created independently next to each other.

    Args:
        n_steps (int): Length of the signal.
        batch_size (int): Number of batches.
        t_min (float): Minimum hold time of an amplitude.
        t_max (float): Maximum hold time of an amplitude
        key (jax.random.PRNGKey): Random key for JAX random sampling.
    """
    actions = []
    for _ in range(batch_size):
        subkey, key = jax.random.split(key)
        actions.append(aprbs_single_batch(n_steps, t_min, t_max, subkey)[..., None])
    return jnp.stack(actions, axis=0)



"""RLS estimator based on the description given in, e.g., ["Data-Driven Recursive Least Squares Estimation
for Model Predictive Current Control of Permanent Magnet Synchronous Motors", A. Brosch et al., 2021] and
["Universal Direct Torque Controller for Permanent Magnet Synchronous Motors via Meta-Reinforcement
Learning", D. Jakobeit et al., 2025].
"""

import jax
import jax.numpy as jnp
import equinox as eqx


class RLS(eqx.Module):
    num_regressors: int
    num_measurements: int
    lambda_: float
    w: jax.Array
    P: jax.Array

    def __init__(
        self,
        num_regressors: int,
        num_measurements: int,
        lambda_: float,
        w: jax.Array | None = None,
        P: jax.Array | None = None,
    ):
        """Creates a basic RLS model. Either from existing w and P arrays or by initializing
        them to zero and identity matrix, if they are not specified.

        Args:
            num_regressors (int): Number of regressors, i.e., number of inputs
            num_measurements (int): Number of measurements, i.e., number of outputs
            lambda_ (float): Forgetting factor, that determines how quickly the parameters
                are overwritten based on new regressors and measurements
            w (jax.Array | None): Weight matrix with shape (num_regressors, num_measurements)
                or None, which defaults to a corresponding zero matrix
            P (jax.Array | None): Covariance matrix with shape (num_regressors, num_regressors)
                or None, which defaults to a corresponding identity matrix
        """
        self.num_regressors = num_regressors
        self.num_measurements = num_measurements
        self.lambda_ = lambda_

        self.w = jnp.zeros((self.num_regressors, self.num_measurements)) if w is None else w
        self.P = jnp.eye(self.num_regressors) if P is None else P

    @eqx.filter_jit
    def predict(self, x: jax.Array) -> jax.Array:
        """Predict the output for a given regressor vector using the RLS model. Uses the parameter
        matrix currently stored in the RLS model.

        Args:
            x (jax.Array): The regressor vector with shape (num_regressors, 1) to use for the prediction

        Returns:
            y_pred (jax.Array): The predicted output vector with shape (num_measurements, 1)
        """
        y_pred = self.w.T @ x
        return y_pred

    @staticmethod
    @eqx.filter_jit
    def update(rls: "RLS", x: jax.Array, d: jax.Array) -> "RLS":
        """Update the parameters of the RLS model.

        NOTE that this is done in a static method that takes the momentary RLS model
        as an input and creates a new instance with new parameters. This is done since
        eqx.Modules are immutable.

        Args:
            rls (RLS): The momentary RLS model to update
            x (jax.Array): The regressor vector with shape (num_regressors, 1)
            d (jax.Array): The measured output value with shape (num_measurements, 1)

        Returns:
            rls (RLS): The updated RLS model
        """
        P = rls.P
        w = rls.w

        c = (P @ x) / (rls.lambda_ + jnp.squeeze(x.T @ P @ x))
        w_new = w + c @ (d - w.T @ x).T
        P_new = (jnp.eye(rls.num_regressors) - c @ x.T) @ P / rls.lambda_

        return RLS(rls.num_regressors, rls.num_measurements, rls.lambda_, w=w_new, P=P_new)


class PMSM_RLS(eqx.Module):
    rls: eqx.Module

    def __init__(self, lambda_: float = 0.99, rls: RLS | None = None):
        """RLS model specifically for the PMSM.

        The num_regressors is set to 5 with the expected regressor vector [i_d, i_q, u_d, u_q, 1].
        The num_measurements is set to 2 with the expected measurement vector [i_d, i_q].

        Args:
            lambda_ (float): Forgetting factor, that determines how quickly the parameters
                are overwritten based on new inputs
            rls (RLS): An existing RLS model or None if a new one should be created
        """
        self.rls = RLS(5, 2, lambda_) if rls is None else rls

    @eqx.filter_jit
    def __call__(self, x: jax.Array) -> jax.Array:
        """Use the underlying RLS model to perform a prediction.

        Args:
            x (jax.Array): The regressor vector with shape (num_regressors, 1)
        """
        return self.rls.predict(x)

    @classmethod
    @eqx.filter_jit
    def update(cls, pmsm_rls: "PMSM_RLS", x: jax.Array, d: jax.Array) -> "PMSM_RLS":
        """Update the PMSM specific RLS model.

        Args:
            pmsm_rls (PMSM_RLS): The momentary PMSM-RLS model
            x (jax.Array): The regressor vector with shape (num_regressors, 1)
            d (jax.Array): The measured output value with shape (num_measurements, 1)

        Returns:
            new_pmsm_rls (PMSM_RLS): The updated PMSM-RLS model
        """

        rls = RLS.update(pmsm_rls.rls, x, d)
        return cls(rls.lambda_, rls)

    @property
    def num_regressors(self) -> int:
        return self.rls.num_regressors

    @property
    def num_measurements(self) -> int:
        return self.rls.num_measurements


class SimulationPMSM_RLS(PMSM_RLS):
    """Wraps the PMSM_RLS model to perform multistep simulation.

    This API is needed to make the RLS model properly usable in DMPE.
    """

    @eqx.filter_jit
    def __call__(self, init_obs: jax.Array, actions: jax.Array, tau=None) -> jax.Array:
        """Simulate multiple steps ahead by applying actions starting from the initial observation.

        Args:
            init_obs (jax.Array): The initial observation, expected to be [i_d, i_q] with shape (2,)
                (will be squeezed in the function)
            actions (jax.Array): The actions to be applied, expected to be a sequence of [u_d, u_q]
                value-pairs. The shape is expected to be (n_sim_steps, 2). The function iterates
                along the first dimension and squeezes the resulting values.
            tau: NOTE This value is not actually used in this version and is only added since it is
                part of the model API for DMPE.

        Returns:
            observations (jax.Array): The simulation / prediction result with shape (n_sim_steps, 2)
        """

        def body_fun(carry, action):
            obs = carry
            rls_in = jnp.concatenate([jnp.squeeze(obs), jnp.squeeze(action), jnp.ones(1)])[..., None]

            obs = super(SimulationPMSM_RLS, self).__call__(rls_in)
            obs = jnp.squeeze(obs)
            return obs, obs

        _, observations = jax.lax.scan(body_fun, jnp.squeeze(init_obs), actions)
        observations = jnp.concatenate([jnp.squeeze(init_obs)[None, :], observations], axis=0)
        return observations

    @classmethod
    @eqx.filter_jit
    def fit(
        cls,
        model: "SimulationPMSM_RLS",
        k: jax.Array,
        observations: jax.Array,
        actions: jax.Array,
    ) -> "SimulationPMSM_RLS":
        """Update the RLS model with the observations and actions at index k (generally this will be
        the most recent interaction with the system).

        Args:
            model (SimulationPMSM_RLS): The momentary SimulationPMSM_RLS model to update
            k (jax.Array): The index of the observations and actions to use for the update
            observations (jax.Array): The observations from the environment, expected to be a sequence
                of [i_d, i_q] value-pairs with shape (N, 2) with N > k
            actions (jax.Array): The actions taken in the environment, expected to be a sequence of
                [u_d, u_q] value-pairs. The shape is expected to be (N-1, 2) with N > k

        Returns:
            model (SimulationPMSM_RLS): The updated SimulationPMSM_RLS model
        """
        obs = jnp.squeeze(observations[k])
        action = jnp.squeeze(actions[k])
        next_obs = jnp.squeeze(observations[k + 1])

        model_in = jnp.concatenate([obs, action, jnp.ones(1)])[..., None]
        model = cls.update(model, x=model_in, d=next_obs[..., None])
        return model



from functools import partial
import json
import argparse
import datetime
from tqdm import tqdm
import pathlib

import jax
import jax.numpy as jnp

from dmpe.algorithms.algorithm_utils import interact_and_observe
from dmpe.evaluation.scripts.pmsm.eval_dmpe import setup_env, TARGETED_DATA_PATH


@partial(jax.jit, static_argnums=(0, 1))
def choose_action(env, penalty_function, proposed_actions, state, choice_key):
    """Choose randomly among the proposed actions that keep the system within bounds for the next step.
    If none of the inputs keep the systems in bounds, apply the one that causes the least penalty.

    This is a heuristic implementation that uses an oracle to ensure compliance with the bounds, but chooses
    mostly randomly among the actions.
    """
    test_obs, _ = jax.vmap(env.step, in_axes=(None, 0, None))(state, proposed_actions, env.env_properties)

    penalty_values = jax.vmap(penalty_function, in_axes=(0, 0))(test_obs, proposed_actions)

    def true_fun(key, data_array, penalty_values):
        """There are no options that keep the system within bounds. Apply the one with the least penalty."""
        idx_min_penalty = jnp.argmin(penalty_values)
        return data_array[idx_min_penalty]

    def false_fun(key, data_array, penalty_values):
        """There are actions that keep the system within bounds. Choose one randomly."""
        valid_points_bool = penalty_values == 0
        prob_points = valid_points_bool.astype(jnp.float32) / jnp.sum(valid_points_bool)
        return jax.random.choice(key, data_array, p=prob_points, axis=0)

    return jax.lax.cond(
        jnp.all(penalty_values != 0), true_fun, false_fun, *(choice_key, proposed_actions, penalty_values)
    )


def run_experiment(rpm, seed):

    print(
        "Running experiment with random walk.",
        f"(seed: {int(seed)}) on the PMSM with {rpm} rpm.",
    )

    # Check that the targeted data folder actually exist:
    results_path = TARGETED_DATA_PATH / pathlib.Path("heuristics") / pathlib.Path("random_walk")
    print(f"Results will be written to: '{results_path}'.")
    assert results_path.exists(), (
        f"The expected results path '{results_path}' does not seem to exist. Please create the necessary file structure "
        + "or adapt the path."
    )

    n_time_steps = 15_000
    n_tries = 5_000

    env, penalty_function = setup_env(rpm)
    obs, state = env.reset(env.env_properties)
    dim_obs_space = obs.shape[0]
    dim_action_space = env.action_dim

    observations = jnp.zeros((n_time_steps, dim_obs_space))
    observations = observations.at[0].set(obs)
    actions = jnp.zeros((n_time_steps - 1, dim_action_space))

    key = jax.random.key(seed)
    key, action_key = jax.random.split(key)

    action = jax.random.normal(action_key, shape=(env.action_dim,))

    for k in tqdm(range(n_time_steps)):

        key, action_key, choice_key = jax.random.split(key, 3)
        proposed_actions = action + jax.random.normal(
            action_key,
            shape=(
                n_tries,
                env.action_dim,
            ),
        )

        action = choose_action(env, penalty_function, proposed_actions, state, choice_key)

        next_obs, next_state, actions, observations = interact_and_observe(
            env=env, k=jnp.array([k]), action=action, state=state, actions=actions, observations=observations
        )

        state = next_state
        obs = next_obs

    # experiment finished, save results
    file_name = datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")

    with open(results_path / pathlib.Path(f"data_rpm_{rpm}_{file_name}.json"), "w") as fp:
        json.dump(dict(observations=observations.tolist(), actions=actions.tolist()), fp)

    jax.clear_caches()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Run random walk on the PMSM environment.")
    parser.add_argument("--gpu_id", type=int, default=0, help="GPU id to use.")
    args = parser.parse_args()

    gpus = jax.devices()
    jax.config.update("jax_default_device", gpus[args.gpu_id])

    rpms = [0, 3000, 5000, 7000, 9000]

    for rpm in rpms:
        seeds = list(range(50, 61))
        for seed in seeds:
            run_experiment(rpm, seed)



import argparse
import pathlib

REPO_ROOT_PATH = pathlib.Path(__file__).parent.parent.parent.parent
DATA_PATH = REPO_ROOT_PATH / pathlib.Path("data")

if __name__ == "__main__":
    print("repo root: ", REPO_ROOT_PATH)

    print(f"Creating 'data' folder at '{DATA_PATH}'")
    DATA_PATH.mkdir(parents=False, exist_ok=True)

    parser = argparse.ArgumentParser(description="Create experiments folder structure.")
    parser.add_argument("--which", type=str, default="all", help="One of 'all', 'classical', or 'pmsm'.")

    args = parser.parse_args()
    which = args.which

    if which == "classical" or which == "all":

        systems = ["fluid_tank", "pendulum", "cart_pole"]
        algorithms = ["dmpe", "igoats", "sgoats", "perfect_model_dmpe"]

        for algo in algorithms:
            for system in systems:
                system_algo_path = (
                    DATA_PATH / pathlib.Path("classical_systems") / pathlib.Path(algo) / pathlib.Path(system)
                )
                system_algo_path.mkdir(parents=True, exist_ok=True)

    if which == "pmsm" or which == "all":
        structure = {
            "dmpe": ["NODE", "PM", "RLS"],
            "igoats": ["_"],
            "heuristics": ["current_plane_sweep", "random_walk"],
        }

        for algo_class, algo_names in structure.items():
            for algo_name in algo_names:
                algo_path = DATA_PATH / pathlib.Path("pmsm") / pathlib.Path(algo_class) / pathlib.Path(algo_name)
                algo_path.mkdir(parents=True, exist_ok=True)

    print("Successfully created all directories.")



from typing import Callable

import jax
import jax.numpy as jnp
import optax


def get_alg_params(consider_action_distribution: bool, penalty_function: Callable):
    """Get parameters for the iGOATS algorithm in the PMSM experiments."""
    h = 4
    a = 4

    alg_params = dict(
        consider_action_distribution=consider_action_distribution,
        prediction_horizon=h,
        application_horizon=a,
        bounds_amplitude=(-1, 1),
        bounds_duration=(1, 50),
        population_size=50,
        n_generations=25,
        featurize=lambda x: x,
        rng=None,
        compress_data=True,
        compression_target_N=500,
        compression_feat_dim=0,
        compression_dist_th=0.1,
        penalty_function=penalty_function,
    )

    return alg_params



import jax
import jax.nn as jnn
import jax.numpy as jnp

import equinox as eqx
import diffrax


class MLP(eqx.Module):
    mlp: eqx.nn.MLP

    def __init__(
        self,
        obs_dim: int,
        action_dim: int,
        width_size: int,
        depth: int,
        *,
        key: jax.random.PRNGKey,
        **kwargs,
    ):
        """Basic state-space MLP. Prediction is based on observation and action.

        Is generally used here to predict the dx/dt in NODE models.

        Args:
            obs_dim (int): Observation dimensionality
            action_dim (int): Action dimensionality
            width_size (int): Layer width (same for all layer) in the MLP
            depth (int): Number of hidden layers in the MLP
            key (jax.random.PRNGKey): Random key for model initialization
        """
        super().__init__(**kwargs)
        self.mlp = eqx.nn.MLP(
            in_size=(obs_dim + action_dim),
            out_size=obs_dim,
            width_size=width_size,
            depth=depth,
            activation=jnn.leaky_relu,
            key=key,
        )

    def __call__(self, obs: jax.Array, action: jax.Array) -> jax.Array:
        obs_action = jnp.hstack([obs, action])
        return self.mlp(obs_action)


class NeuralODE(eqx.Module):
    func: MLP
    _solver: diffrax.AbstractSolver

    def __init__(
        self,
        solver: diffrax.AbstractSolver,
        obs_dim: int,
        action_dim: int,
        width_size: int,
        depth: int,
        *,
        key: jax.random.PRNGKey,
        **kwargs,
    ):
        """Neural ordinary differential equation implementation.

        Args:
            solver (diffrax.AbstractSolver): The ODE solver to use for the Neural ODE.
            obs_dim (int): Observation dimensionality
            action_dim (int): Action dimensionality
            width_size (int): Layer width (same for all layer) in the MLP
            depth (int): Number of hidden layers in the MLP
            key (jax.random.PRNGKey): Random key for model initialization
        """

        super().__init__(**kwargs)
        self.func = MLP(obs_dim, action_dim, width_size, depth, key=key)
        self._solver = solver

    def __call__(self, init_obs: jax.Array, actions: jax.Array, tau: float) -> jax.Array:
        """Simulate the observation trajectory using the ODE solver.

        Args:
            init_obs (jax.Array): Initial observation to start the simulation from
            actions (jax.Array): Actions / inputs to be applied
            tau (float): Time step size for the simulation
        """

        args = (actions, None)

        def action_helper(t, args):
            actions = args
            return actions[jnp.array(t / tau, int)]

        def vector_field(t, y, args):
            actions, _ = args

            action = action_helper(t, actions)
            dy_dt = self.func(y, action)
            return tuple(dy_dt)

        term = diffrax.ODETerm(vector_field)
        t0 = 0
        t1 = tau * actions.shape[0]

        y0 = tuple(init_obs)
        saveat = diffrax.SaveAt(ts=jnp.linspace(t0, t1, 1 + int(t1 / tau)))
        solution = diffrax.diffeqsolve(term, self._solver, t0, t1, dt0=tau, y0=y0, args=args, saveat=saveat)

        return jnp.transpose(jnp.array(solution.ys))


class NeuralODEPendulum(NeuralODE):
    """Pendulum specific model that deals with the periodic properties of the angle information."""

    def __call__(self, init_obs: jax.Array, actions: jax.Array, tau: float) -> jax.Array:
        observations = super().__call__(init_obs, actions, tau)
        return jnp.stack([(((observations[..., 0] + 1) % 2) - 1), observations[..., 1]], axis=-1)


class NeuralEulerODE(eqx.Module):
    func: MLP

    def __init__(
        self,
        obs_dim: int,
        action_dim: int,
        width_size: int,
        depth: int,
        *,
        key: jax.random.PRNGKey,
        **kwargs,
    ):
        """Neural ODE model with Euler as the ODE solver.

        Very fast to use compared to the models based on more complex ODE solvers.
        However, will generally be less accurate. In essence this model boils down
        to a difference equation model instead of a true NODE.

        Args:
            solver (diffrax.AbstractSolver): The ODE solver to use for the Neural ODE.
            obs_dim (int): Observation dimensionality
            action_dim (int): Action dimensionality
            width_size (int): Layer width (same for all layer) in the MLP
            depth (int): Number of hidden layers in the MLP
            key (jax.random.PRNGKey): Random key for model initialization
        """
        super().__init__(**kwargs)
        self.func = MLP(obs_dim, action_dim, width_size, depth, key=key)

    def step(self, obs: jax.Array, action: jax.Array, tau: float) -> jax.Array:
        """Takes a single step in the simulation using the Euler method.

        Args:
            obs (jax.Array): Momentary observation
            action (jax.Array): Action to be applied
            tau (float): Time step size for the simulation

        Returns:
            next_obs (jax.Array): The predicted next observation.
        """
        next_obs = obs + tau * self.func(obs, action)
        return next_obs

    def __call__(self, init_obs: jax.Array, actions: jax.Array, tau: float) -> jax.Array:
        """Simulate the observation trajectory using the Euler ODE solver.

        Args:
            init_obs (jax.Array): Initial observation to start the simulation from
            actions (jax.Array): Actions / inputs to be applied
            tau (float): Time step size for the simulation
        """

        def body_fun(carry, action):
            obs = carry
            obs = self.step(obs, action, tau)
            return obs, obs

        _, observations = jax.lax.scan(body_fun, init_obs, actions)
        observations = jnp.concatenate([init_obs[None, :], observations], axis=0)
        return observations


class NeuralEulerODEPendulum(NeuralEulerODE):
    """Pendulum specific model that deals with the periodic properties of the angle information."""

    def step(self, obs: jax.Array, action: jax.Array, tau: float) -> jax.Array:
        next_obs = super().step(obs, action, tau)
        next_obs = jnp.stack([(((next_obs[..., 0] + 1) % 2) - 1), next_obs[..., 1]], axis=-1)
        return next_obs


class NeuralEulerODECartpole(NeuralEulerODE):
    """Cartpole specific model that deals with the periodic properties of the angle information."""

    def step(self, obs: jax.Array, action: jax.Array, tau: float) -> jax.Array:
        next_obs = super().step(obs, action, tau)
        next_obs = jnp.stack(
            [next_obs[..., 0], next_obs[..., 1], (((next_obs[..., 2] + 1) % 2) - 1), next_obs[..., 3]], axis=-1
        )
        return next_obs


class NeuralEulerODEPMSM(NeuralEulerODE):
    """PMSM-specific NODE model that roughly scales the output to fit the normal output range more accurately."""

    def step(self, obs: jax.Array, action: jax.Array, tau: float) -> jax.Array:
        next_obs = obs + tau * self.func(obs, action) * 1e4
        return next_obs



from typing import Callable

import jax
import jax.numpy as jnp
import equinox as eqx
import optax

from dmpe.models.model_utils import simulate_ahead


@eqx.filter_jit
@eqx.filter_grad
def grad_loss(
    model: eqx.Module,
    true_obs: jax.Array,
    actions: jax.Array,
    tau: float,
    featurize: Callable,
) -> jax.Array:
    """Basic MSE implementation for model training.

    The model predicts the observation sequence based on the actions to be applied
    and the first element in the sequence. The predicted observations are then
    compared with the ground truth values.

    Note: The function itself actually returns the MSE value. Due to the gradient
    decorator the function returns the gradient of the MSE with respect to the
    model parameters instead. This is also why the function is called grad_loss
    and not loss.

    Args:
        model (eqx.Module): The model to be trained.
        true_obs (jax.Array): A batched sequence of ground-truth observations with
            shape (batch_size, sequence_length, obs_dim)
        actions (jax.Array): The actions that have been applied to generate the observation
            sequence.
        tau (float): Sampling frequency of the system.
        featurize (Callable): Function that potentially adds additional features to the
            observations. For instance angle information is usually transformed to sin
            and cos of the angle to have a better grasp on similar angles.

    Returns:
        The resulting MSE loss value (jax.Array)
        (Due to the gradient decorator this actually returns the gradient of the MSE w.r.t. the model parameters)
    """
    pred_obs = jax.vmap(simulate_ahead, in_axes=(None, 0, 0, None))(model, true_obs[:, 0, :], actions, tau)

    feat_pred_obs = jax.vmap(featurize, in_axes=(0))(pred_obs)
    feat_true_obs = jax.vmap(featurize, in_axes=(0))(true_obs)

    return jnp.mean((feat_pred_obs - feat_true_obs) ** 2)


@eqx.filter_jit
def make_step(
    model: eqx.Module,
    observations: jax.Array,
    actions: jax.Array,
    tau: float,
    opt_state: optax.OptState,
    featurize: Callable,
    optim: optax.GradientTransformation,
) -> tuple[eqx.Module, optax.OptState]:
    """Performs a single update step on the model.

    First the gradient of the MSE loss with respect to the model parameters is computed.
    Then the optimizers tunes the model parameters based on this.

    Args:
        model (eqx.Module): The model to be trained.
        observations (jax.Array): A batched sequence of ground-truth observations with
            shape (batch_size, sequence_length, obs_dim)
        actions (jax.Array): The actions that have been applied to generate the observation
            sequence.
        tau (float): Sampling frequency of the system.
        opt_state (optax.OptState): State of the model optimizer
        featurize (Callable): Function that potentially adds additional features to the
            observations. For instance angle information is usually transformed to sin
            and cos of the angle to have a better grasp on similar angles.
        optim (optax.GradientTransformation): Model optimizer

    Returns:
        model (eqx.Module): The updated model
        opt_state (optax.OptState): The updated state of the model optimizer
    """
    grads = grad_loss(model, observations, actions, tau, featurize)
    updates, opt_state = optim.update(grads, opt_state)
    model = eqx.apply_updates(model, updates)
    return model, opt_state


@eqx.filter_jit
def load_single_batch(
    observations_array: jax.Array,
    actions_array: jax.Array,
    starting_points: jax.Array,
    sequence_length: int,
) -> tuple[jax.Array, jax.Array]:
    """Loads a single batch of data from the memory arrays.

    Args:
        observations_array (jax.Array): An array containing all gathered observations. Note
            that the array is zero-padded to always have the same length. The shape is always
            (max_length, obs_dim) regardless of how many observations < max_length have been
            gathered.
        actions_array (jax.Array): Array with actions corresponding to the observations array.
        starting_points (jax.Array): The starting points for all batches with shape (batch_size,)
        sequence_length (int): The length of the sequences in each batch. The minimum sensible
            sequence length is 2. As the actions array is empty otherwise and there is only
            a single observation, which one cannot properly learn with.

    Returns:
        batched_observations (jax.Array): A batch of observations with shape (batch_size, sequence_length, obs_dim)
        batched_actions (jax.Array): A batch of actions with shape (batch_size, sequence_length-1, action_dim)
    """

    slice = jnp.linspace(
        start=starting_points, stop=starting_points + sequence_length, num=sequence_length, dtype=int
    ).T

    batched_observations = observations_array[slice]
    batched_actions = actions_array[slice]

    batched_observations = batched_observations[:, :, :]
    batched_actions = batched_actions[:, :-1, :]
    return batched_observations, batched_actions


@eqx.filter_jit
def precompute_starting_points(
    n_train_steps: int,
    k: jax.Array,
    sequence_length: int,
    training_batch_size: int,
    loader_key: jax.random.PRNGKey,
) -> tuple[jax.Array, jax.random.PRNGKey]:
    """Sample where the batch sequences should start in the gathered data arrays.

    The maximum possible starting index is k + 1 - sequence_length as beyond that
    the data arrays are filled with zeros.

    Args:
        n_train_steps (int): Number of training steps to be performed.
        k (jax.Array): Denotes to which index the memory arrays are currently filled.
        sequence_length (int): The length of the sequences in each batch. The minimum sensible
            sequence length is 2. As the actions array is empty otherwise and there is only
            a single observation, which one cannot properly learn with.
        training_batch_size (int): The size of the training batches.
        loader_key (jax.random.PRNGKey): Random key for sampling.

    Returns:
        starting_points (jax.Array): The starting points for all batches with shape (n_train_steps, training_batch_size)
        next_loader_key (jax.random.PRNGKey): The updated random key.
    """
    next_loader_key, momentary_loader_key = jax.random.split(loader_key, 2)

    index_normalized = jax.random.uniform(momentary_loader_key, shape=(n_train_steps, training_batch_size)) * (
        k + 1 - sequence_length
    )
    starting_points = index_normalized.astype(jnp.int32)

    return starting_points, next_loader_key


@eqx.filter_jit
def fit(
    model: eqx.Module,
    n_train_steps: int,
    starting_points: jax.Array,
    sequence_length: int,
    observations: jax.Array,
    actions: jax.Array,
    tau: float,
    featurize: Callable,
    optim: optax.GradientTransformation,
    init_opt_state: optax.OptState,
) -> tuple[eqx.Module, optax.OptState]:
    """Fit the model on the gathered data.

    For 'n_train_steps' iterations, a batch of observations and actions is drawn
    according to the corresponding line in the starting_points array. In each iteration
    a training step is performed based on the drawn batches of data.

    Args:
        model (eqx.Module): The model to be trained.
        n_train_steps (int): Number of consecutive training steps to performed.
        starting_points (jax.Array): The starting points for all batches with shape
            (n_train_steps, training_batch_size)
        sequence_length (int): The length of the sequences in each batch. The minimum sensible
            sequence length is 2. As the actions array is empty otherwise and there is only
            a single observation, which one cannot properly learn with.
        observations (jax.Array): An array containing all gathered observations. Note
            that the array is zero-padded to always have the same length. The shape is always
            (max_length, obs_dim) regardless of how many observations < max_length have been
            gathered.
        actions (jax.Array): Array with actions corresponding to the observations array.
        tau (float): Sampling frequency of the system.
        featurize (Callable): Function that potentially adds additional features to the
            observations. For instance angle information is usually transformed to sin
            and cos of the angle to have a better grasp on similar angles.
        optim (optax.GradientTransformation): Model optimizer
        init_opt_state (optax.OptState): State of the model optimizer

    Returns:
        model (eqx.Module): The updated model
        opt_state (optax.OptState): The updated state of the model optimizer
    """

    dynamic_init_model_state, static_model_state = eqx.partition(model, eqx.is_array)
    init_carry = (dynamic_init_model_state, init_opt_state)

    def body_fun(i, carry):
        dynamic_model_state, opt_state = carry
        model_state = eqx.combine(static_model_state, dynamic_model_state)

        batched_observations, batched_actions = load_single_batch(
            observations, actions, starting_points[i, ...], sequence_length
        )
        new_model_state, new_opt_state = make_step(
            model_state, batched_observations, batched_actions, tau, opt_state, featurize, optim
        )

        new_dynamic_model_state, new_static_model_state = eqx.partition(new_model_state, eqx.is_array)
        assert eqx.tree_equal(static_model_state, new_static_model_state) is True
        return (new_dynamic_model_state, new_opt_state)

    final_dynamic_model_state, final_opt_state = jax.lax.fori_loop(
        lower=0, upper=n_train_steps, body_fun=body_fun, init_val=init_carry
    )
    final_model = eqx.combine(static_model_state, final_dynamic_model_state)
    return final_model, final_opt_state


class ModelTrainer(eqx.Module):
    """A class that carries the necessary tools for training simulation models based on eqx.Modules.

    Args:
        start_learning (int): How many steps the trainer should wait before starting the training.
            This is done so that a minimum of data can be gathered before training is started.
        training_batch_size (int): The size of the training batches.
        n_train_steps (int): Number of consecutive training steps to performed.
        sequence_length (int): The length of the sequences in each batch. The minimum sensible
            sequence length is 2. As the actions array is empty otherwise and there is only
            a single observation, which one cannot properly learn with.
        featurize (Callable): Function that potentially adds additional features to the observations.
        model_optimizer (optax.GradientTransformation): Model optimizer
        tau (float): Sampling frequency of the system.

    """

    start_learning: int
    training_batch_size: int
    n_train_steps: int
    sequence_length: int
    featurize: Callable
    model_optimizer: optax.GradientTransformation
    tau: float

    @eqx.filter_jit
    def fit(
        self,
        model: eqx.Module,
        k: jax.Array,
        observations: jax.Array,
        actions: jax.Array,
        opt_state: optax.OptState,
        loader_key: jax.random.PRNGKey,
    ) -> tuple[eqx.Module, optax.OptState, jax.random.PRNGKey]:
        """Sample batches and train the model based on the data gathered up to this point.

        Args:
            model (eqx.Module): The model to be trained.
            k (jax.Array): Denotes to which index the memory arrays are currently filled.
            observations (jax.Array): An array containing all gathered observations. Note
                that the array is zero-padded to always have the same length. The shape is always
                (max_length, obs_dim) regardless of how many observations < max_length have been
                gathered.
            actions (jax.Array): Array with actions corresponding to the observations array.
            opt_state (optax.OptState): State of the model optimizer
            loader_key (jax.random.PRNGKey): Random key for sampling.
        Returns:
            final_model (eqx.Module): The updated model
            final_opt_state (optax.OptState): The updated state of the model optimizer
            next_loader_key (jax.random.PRNGKey): The updated random key.
        """
        starting_points, loader_key = precompute_starting_points(
            n_train_steps=self.n_train_steps,
            k=k,
            sequence_length=self.sequence_length,
            training_batch_size=self.training_batch_size,
            loader_key=loader_key,
        )

        final_model, final_opt_state = fit(
            model=model,
            n_train_steps=self.n_train_steps,
            starting_points=starting_points,
            sequence_length=self.sequence_length,
            observations=observations,
            actions=actions,
            tau=self.tau,
            featurize=self.featurize,
            optim=self.model_optimizer,
            init_opt_state=opt_state,
        )
        return final_model, final_opt_state, loader_key



from typing import Callable
import optax

from dmpe.models.models import NeuralEulerODEPMSM
from dmpe.models.rls import SimulationPMSM_RLS
from dmpe.utils.density_estimation import get_uniform_target_distribution


def get_alg_params(consider_action_distribution: bool, penalty_function: Callable):
    """Get parameters for the DMPE algorithm in the PMSM experiments."""

    alg_params = dict(
        bandwidth=0.08,
        n_prediction_steps=5,
        points_per_dim=21,
        grid_extend=1.05,
        excitation_optimizer=optax.adabelief(1e-2),
        n_opt_steps=200,
        start_optimizing=5,
        consider_action_distribution=consider_action_distribution,
        penalty_function=penalty_function,
        target_distribution=None,
        clip_action=False,
        n_starts=10,
        reuse_proposed_actions=True,
    )

    alg_params["target_distribution"] = get_uniform_target_distribution(
        dim=4 if consider_action_distribution else 2,
        points_per_dim=alg_params["points_per_dim"],
        bandwidth=alg_params["bandwidth"],
        grid_extend=alg_params["grid_extend"],
        consider_action_distribution=consider_action_distribution,
        penalty_function=penalty_function,
    )
    return alg_params


def get_RLS_params(consider_action_distribution, penalty_function):
    """Get parameters for the application of RLS models for the DMPE pmsm experiments."""
    alg_params = get_alg_params(
        consider_action_distribution=consider_action_distribution, penalty_function=penalty_function
    )

    model_params = dict(lambda_=0.9)
    model_trainer_params = None
    model_class = SimulationPMSM_RLS

    return alg_params, model_params, model_class, model_trainer_params


def get_NODE_params(consider_action_distribution, penalty_function):
    """Get parameters for the application of NODE models for the DMPE pmsm experiments."""

    alg_params = get_alg_params(
        consider_action_distribution=consider_action_distribution, penalty_function=penalty_function
    )

    model_params = dict(obs_dim=2, action_dim=2, width_size=64, depth=3, key=None)
    model_trainer_params = dict(
        start_learning=alg_params["n_prediction_steps"],
        training_batch_size=64,
        n_train_steps=5,
        sequence_length=alg_params["n_prediction_steps"],
        featurize=lambda x: x,
        model_lr=1e-4,
    )
    model_class = NeuralEulerODEPMSM

    return alg_params, model_params, model_class, model_trainer_params


def get_PM_params(consider_action_distribution, penalty_function):
    alg_params = get_alg_params(
        consider_action_distribution=consider_action_distribution, penalty_function=penalty_function
    )

    alg_params["n_opt_steps"] = 100

    model_params = None
    model_trainer_params = None
    model_class = None

    return alg_params, model_params, model_class, model_trainer_params



from typing import Callable

import jax
import jax.numpy as jnp

import equinox as eqx


def select_bandwidth(
    delta_z: float,
    dim: int,
    n_g: int,
    percentage: float,
) -> float:
    """Select a bandwidth for the kernel density estimate by a rough heuristic.

    The bandwidth is designed so that the kernel is still at a given percentage of
    its maximum value when a step is taken in each dimension of the underlying
    grid.

    Args:
        delta_z (float): The size of the space in each dimension. Assumed to be
            symmetrical
        dim (int): The dimension of the space.
        n_g (int): Number of grid points per dimension.
        percentage (float): The percentage of the maximum value of the kernel at
            the other grid point reached by stepping once in each dimension on
            the grid.
    Returns:
        bandwidth (float): The resulting proposed bandwidth
    """
    return delta_z * jnp.sqrt(dim) / (n_g * jnp.sqrt(-2 * jnp.log(percentage)))


@jax.jit
def gaussian_kernel(z: jax.Array, bandwidth: float) -> jax.Array:
    """Evaluates the Gaussian RBF kernel at z with given bandwidth. This can take arbitrary
    dimensions for 'z' and will compute the output by broadcasting. The last dimension of
    the input needs to be the dimension of the data which is reduced along.

    Args:
        z (jax.Array): The input data points with shape (..., data_dim). The first dimensions
            are broadcast over and the last dimension is assumed to be the feature/data-dimension.
        bandwidth (float): Bandwidth of the kernel. Always symmetrical in the feature space.

    Returns:
        The kernel value as a jax.Array with shape (..., 1). Where all dimensions before the
        last one are kept as they are.
    """
    data_dim = z.shape[-1]
    factor = bandwidth**data_dim * jnp.power(2 * jnp.pi, data_dim / 2)
    return 1 / factor * jnp.exp(-jnp.linalg.norm(z, axis=-1) ** 2 / (2 * bandwidth**2))


class DensityEstimate(eqx.Module):
    """Holds an estimate of the data density of sampled data points.

    Args:
        p (jax.Array): The probability estimates at the grid points
        z_g (jax.Array): The grid points to which the probability estimates belong
        bandwidth (jax.Array): The bandwidth of the kernel density estimate
        n_observations (jax.Array): The number of observations that make up the momentary
            estimate
    """

    p: jax.Array
    z_g: jax.Array
    bandwidth: jax.Array
    n_observations: jax.Array

    @classmethod
    def from_estimate(
        cls,
        p: jax.Array,
        n_additional_observations: int,
        density_estimate: "DensityEstimate",
    ) -> "DensityEstimate":
        """Create a density estimate from an existing estimate.

        The computation is usually that p is constructed from the DensityEstimate.p before the
        update together with the kernel density estimation over new data points. The result of
        this computation is the input 'p'.

        Args:
            p (jax.Array): The new probability estimate
            n_additional_observations (int): The number of data points that have been added
                in the update
            density_estimate (DensityEstimate): The density estimate before the update

        Returns:
            The updated density estimate (DensityEstimate)
        """

        return cls(
            p=p,
            n_observations=(density_estimate.n_observations + n_additional_observations),
            z_g=density_estimate.z_g,
            bandwidth=density_estimate.bandwidth,
        )

    @classmethod
    def from_observations_actions(
        cls,
        observations: jax.Array,
        actions: jax.Array,
        use_actions: bool = True,
        points_per_dim: int = 30,
        z_min: float = -1,
        z_max: float = 1,
        bandwidth: float = 0.05,
    ) -> "DensityEstimate":
        """Create a fresh density estimate from gathered in the form of observations and actions.

        It is assumed that the feature vector is made up from observations and actions together
        when 'use_actions' is True and only from the observations if it is False.

        Often, the actions will have one less element compared to the observations. This occurs
        when an observation has resulted from the application of the last action but a new action
        has not been chosen yet. The last observation is then omitted from the density estimate.

        Args:
            observations (jax.Array): The observations to be part of the density estimate
            actions (jax.Array): The actions to be part of the density estimate
            use_actions (bool): A flag indicating whether the actions are to be part of the density
                estimate
            points_per_dim (int): The number of grid points per dimension. Always identical for each
                dimension
            z_min (float): The minimum value of the grid in each dimension
            z_max (float): The maximum value of the grid in each dimension
            bandwidth (float): The bandwidth of the kernel density estimate

        Returns:
            The resulting density estimate (DensityEstimate)
        """

        if observations.shape[0] == actions.shape[0] + 1:
            data_points = (
                jnp.concatenate([observations[0:-1, :], actions], axis=-1)[None] if use_actions else observations[None]
            )
        else:
            data_points = jnp.concatenate([observations, actions], axis=-1)[None] if use_actions else observations[None]

        return cls.from_dataset(data_points, points_per_dim, z_min, z_max, bandwidth)

    @classmethod
    def from_dataset(
        cls,
        data_points: jax.Array,
        points_per_dim: int = 30,
        z_min: float = -1,
        z_max: float = 1,
        bandwidth: float = 0.05,
    ) -> "DensityEstimate":
        """Create a fresh density estimate from gathered in the form of a sequence of feature vectors.

        Args:
            data_points (jax.Array): The data points to be part of the density estimate
            points_per_dim (int): The number of grid points per dimension. Always identical for each
                dimension
            z_min (float): The minimum value of the grid in each dimension
            z_max (float): The maximum value of the grid in each dimension
            bandwidth (float): The bandwidth of the kernel density estimate

        Returns:
            The resulting density estimate (DensityEstimate)
        """
        dim = data_points.shape[-1]

        n_grid_points = points_per_dim**dim
        density_estimate = cls(
            p=jnp.zeros([1, n_grid_points, 1]),
            z_g=build_grid(dim, z_min, z_max, points_per_dim),
            bandwidth=jnp.array([bandwidth]),
            n_observations=jnp.array([0]),
        )
        density_estimate = jax.vmap(
            update_density_estimate_multiple_observations,
            in_axes=(DensityEstimate(0, None, None, None), 0),
            out_axes=(DensityEstimate(0, None, None, None)),
        )(
            density_estimate,
            data_points,
        )
        return density_estimate


@jax.jit
def update_density_estimate_single_observation(
    density_estimate: DensityEstimate,
    data_point: jax.Array,
) -> DensityEstimate:
    """Recursive update to the kernel density estimation (KDE) on a fixed grid.

    Args:
        density_estimate (DensityEstimate): The density estimate before the update
        data_point (jax.Array): The new data point

    Returns:
        The updated density estimate (DensityEstimate)
    """
    kernel_value = gaussian_kernel(z=density_estimate.z_g - data_point, bandwidth=density_estimate.bandwidth)
    p_est = (
        1
        / (density_estimate.n_observations + 1)
        * (density_estimate.n_observations * density_estimate.p + kernel_value[..., None])
    )

    return DensityEstimate.from_estimate(p=p_est, n_additional_observations=1, density_estimate=density_estimate)


@jax.jit
def update_density_estimate_multiple_observations(
    density_estimate: DensityEstimate,
    data_points: jax.Array,
) -> DensityEstimate:
    """Add a new sequence of data points to the current data density estimate.

    Args:
        density_estimate (DensityEstimate): The density estimate before the update
        data_points (jax.Array): The sequence of data_points

    Returns:
        The updated density estimate (DensityEstimate)
    """

    def shifted_gaussian_kernel(z, data_points, bandwidth):
        # created to enable vmapping of data_points without vmapping of z
        return gaussian_kernel(z - data_points, bandwidth)

    new_sum_part = jax.vmap(shifted_gaussian_kernel, in_axes=(None, 0, None))(
        density_estimate.z_g, data_points, density_estimate.bandwidth
    )
    new_sum_part = jnp.sum(new_sum_part, axis=0)[..., None]
    p_est = (
        1
        / (density_estimate.n_observations + data_points.shape[0])
        * (density_estimate.n_observations * density_estimate.p + new_sum_part)
    )

    return DensityEstimate.from_estimate(
        p=p_est, n_additional_observations=data_points.shape[0], density_estimate=density_estimate
    )


def build_grid(dim: int, low: float, high: float, points_per_dim: int) -> jax.Array:
    """Build a uniform grid of points in the given dimension.

    Args:
        dim (int): Dimensionality of the grid (and feature vector z)
        low (float): The minimum value of the grid in each dimension
        high (float): The maximum value of the grid in each dimension
        points_per_dim (int): The number of grid points per dimension. Always identical for each
            dimension

    Returns:
        The flattened grid as a jax.Array with shape (points_per_dim**dim, dim)
    """
    xs = [jnp.linspace(low, high, points_per_dim) for _ in range(dim)]

    z_g = jnp.meshgrid(*xs)
    z_g = jnp.stack([_x for _x in z_g], axis=-1)
    z_g = z_g.reshape(-1, dim)

    assert z_g.shape[0] == points_per_dim**dim
    return z_g


def build_grid_2d(low: float, high: float, points_per_dim: int):
    """Shorthand for a uniform 2d grid.

    Args:
        low (float): The minimum value of the grid in each dimension
        high (float): The maximum value of the grid in each dimension
        points_per_dim (int): The number of grid points per dimension. Always identical for each
            dimension

    Returns:
        The flattened grid as a jax.Array with shape (points_per_dim**2, 2)
    """
    return build_grid(2, low, high, points_per_dim)


def build_grid_3d(low: float, high: float, points_per_dim: int):
    """Shorthand for a uniform 3d grid.

    Args:
        low (float): The minimum value of the grid in each dimension
        high (float): The maximum value of the grid in each dimension
        points_per_dim (int): The number of grid points per dimension. Always identical for each
            dimension

    Returns:
        The flattened grid as a jax.Array with shape (points_per_dim**3, 3)
    """
    return build_grid(3, low, high, points_per_dim)


def get_uniform_target_distribution(
    dim: int,
    points_per_dim: int,
    bandwidth: float,
    grid_extend: float,
    consider_action_distribution: bool,
    penalty_function: Callable,
) -> jax.Array:
    """Get a uniform target distribution for the DMPE algorithm based on the grid parameters
    and a penalty function. Only values that are not penalized by the penalty function
    are targeted in the target distribution, but all of them are to be covered uniformly.

    Args:
        dim (int): Number of dimensions for the grid.
        points_per_dim (int): The number of grid points per dimension. Always identical for each
            dimension
        bandwidth (float): The bandwidth of the kernel density estimate
        grid_extend (float): The extent of the grid in each dimension
        consider_action_distribution (bool): A flag indicating whether the action distribution
            is to be considered
        penalty_function (Callable): The penalty function that is used to determine the
            valid grid points

    Returns:
        The target distribution as a jax.Array with shape (points_per_dim**dim, 1)

    """
    z_g = build_grid(dim, low=-grid_extend, high=grid_extend, points_per_dim=points_per_dim)

    if consider_action_distribution:
        constr_func = lambda z_g: penalty_function(z_g[..., None, :2], z_g[..., None, 2:])
    else:
        constr_func = lambda z_g: penalty_function(z_g[..., None, :2], None)

    valid_grid_point = jax.vmap(constr_func, in_axes=0)(z_g) == 0
    constrained_data_points = z_g[jnp.where(valid_grid_point == True)]

    target_distribution = DensityEstimate.from_dataset(
        constrained_data_points[None],
        z_min=-grid_extend,
        z_max=grid_extend,
        points_per_dim=points_per_dim,
        bandwidth=bandwidth,
    )
    return target_distribution.p[0] / jnp.sum(target_distribution.p[0])
